# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The entry interface for algorithms."""

from typing import Callable, Dict

from neural_compressor.common.base_config import BaseConfig
from neural_compressor.common.utils import SMOOTH_QUANT, STATIC_QUANT
from neural_compressor.tensorflow.algorithms import KerasAdaptor, Tensorflow_ITEXAdaptor, TensorFlowAdaptor
from neural_compressor.tensorflow.quantization.config import SmoothQuantConfig
from neural_compressor.tensorflow.utils import BaseModel, KerasModel, TFConfig, register_algo, valid_keras_format


@register_algo(name=STATIC_QUANT)
def static_quant_entry(
    model: BaseModel,
    quant_config: BaseConfig,
    calib_dataloader: Callable = None,
    calib_iteration: int = 100,
    calib_func: Callable = None,
):
    """The main entry to apply static quantization.

    Args:
        model: a fp32 model to be quantized.
        quant_config: a quantization configuration.
        calib_dataloader: a data loader for calibration.
        calib_iteration: the iteration of calibration.
        calib_func: the function used for calibration, should be a substitution for calib_dataloader
        when the built-in calibration function of INC does not work for model inference.

    Returns:
        q_model: the quantized model.
    """
    if isinstance(model, KerasModel):
        assert valid_keras_format(model.model), "Only Sequential or Functional models are supported now."
        framework = KerasAdaptor
    elif TFConfig.global_config["backend"] == "itex":
        framework = Tensorflow_ITEXAdaptor
    else:
        framework = TensorFlowAdaptor

    quantizer = framework(TFConfig.global_config)
    q_model = quantizer.quantize(quant_config, model, calib_dataloader, calib_iteration, calib_func)
    TFConfig.reset_global_config()

    return q_model


@register_algo(name=SMOOTH_QUANT)
def smooth_quant_entry(
    model: BaseModel,
    smooth_quant_config: SmoothQuantConfig,
    calib_dataloader: Callable = None,
    calib_iteration: int = 100,
    calib_func: Callable = None,
):
    """The main entry to apply smooth quantization.

    Args:
        model: a fp32 model to be quantized.
        quant_config: a quantization configuration.
        calib_dataloader: a data loader for calibration.
        calib_iteration: the iteration of calibration.
        calib_func: the function used for calibration, should be a substitution for calib_dataloader
        when the built-in calibration function of INC does not work for model inference.

    Returns:
        q_model: the quantized model.
    """
    assert not isinstance(model, KerasModel), "INC don't support smooth quantization for Keras models now."

    from neural_compressor.tensorflow.algorithms import SmoothQuant

    converter = SmoothQuant(smooth_quant_config, calib_dataloader, calib_iteration, calib_func)
    sq_model = converter(model)

    return sq_model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize Conv2DBackpropInput and Conv3DBackpropInputV2."""

import numpy as np
import tensorflow as tf
from tensorflow.core.framework import graph_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper

from ..quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithDeconv2d(QuantizeNodeBase):
    """Quantize Conv2DBackpropInput and Conv3DBackpropInputV2 and apply the fusion."""

    exclude_deconv_nodes = []

    def __init__(self, **kwargs):
        """Initialization."""
        super().__init__(**kwargs)
        self.sorted_patterns = sorted(self.patterns, key=lambda i: len(i), reverse=True)
        if self.new_api:
            self.fusion_mapping = {
                "DequantizeConv2DBackpropInputQuantizeV2": self.apply_single_deconv2d_fusion,
                "DequantizeConv2DBackpropInputBiasAddQuantizeV2": self.apply_deconv2d_biasadd_fusion,
                "DequantizeConv3DBackpropInputV2QuantizeV2": self.apply_single_deconv3d_fusion,
                "DequantizeConv3DBackpropInputV2BiasAddQuantizeV2": self.apply_deconv3d_biasadd_fusion,
            }

    def apply_single_deconv2d_fusion(self, match_node_name):
        """Apply single Conv2DBackpropInput fusion.

        Dequantize + Conv2DBackpropInput + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]

        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        _, q_inputs = self._get_node_input(normal_inputs[2])
        _, q_weights_inputs = self._get_node_input(normal_inputs[1])

        quantizev2_weights_name = q_weights_inputs[0]
        _, weights_name = self._get_node_input(quantizev2_weights_name)
        weights_min_name = weights_name[1]
        weights_max_name = weights_name[2]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weights_name[0]].node, self.per_channel
        )

        all_input_names = [normal_inputs[0]] + [q_weights_name] + q_inputs[:1]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        all_input_names += q_inputs[1:]
        skip_node_name.append(normal_inputs[2])
        skip_node_name.append(normal_inputs[1])
        skip_node_name.append(weights_name[0])
        skip_node_name.append(weights_min_name)
        skip_node_name.append(weights_max_name)
        skip_node_name.append(quantizev2_weights_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_deconv"

                quantized_node_input_names = all_input_names[:2] + all_input_names[2:] + control_inputs

                node_op = "_FusedQuantizedDeconv2D"
                quantized_deconv_node = helper.create_node(node_op, quantized_node_name, quantized_node_input_names)

                helper.copy_attr(quantized_deconv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_deconv_node, "padding", node.attr["padding"])
                helper.copy_attr(quantized_deconv_node, "data_format", node.attr["data_format"])
                if "explicit_paddings" in node.attr:
                    helper.copy_attr(quantized_deconv_node, "explicit_paddings", node.attr["explicit_paddings"])
                helper.copy_attr(quantized_deconv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_deconv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_deconv_node, "Tfilter", dtypes.qint8)
                # helper.set_attr_string(quantized_conv_node, '_kernel', b'QuantizedMklOp')
                helper.set_attr_dtype(quantized_deconv_node, "out_type", dtypes.qint32)
                # helper.set_attr_dtype(quantized_conv_node, "alpha", dtypes.quint8)

                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_inputs",
                    [
                        dtypes.int32.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        input_data_type.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_outputs",
                    [
                        dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )

                self.add_output_graph_node(quantized_deconv_node)
                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8, False)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, match_node_name[1], dtypes.qint8)

            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_deconv2d_biasadd_fusion(self, match_node_name):
        """Apply Conv2DBackpropInput BiasAdd fusion.

        Dequantize + Conv2DBackpropInput + Biasadd + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]

        second_node = self.node_name_mapping[match_node_name[2]].node
        need_insert_dummy_biasadd = 1
        add_a_node_name = helper.node_name_from_input(second_node.input[0])
        add_a_node = self.node_name_mapping[add_a_node_name].node
        add_b_node_name = helper.node_name_from_input(second_node.input[1])
        add_b_node = self.node_name_mapping[add_b_node_name].node

        # if add_a_node.op != 'Const' and add_b_node.op == 'Const':
        #      need_insert_dummy_biasadd = 0
        # if need_insert_dummy_biasadd:
        #      new_match_node_name = self._insert_dummy_biasadd(match_node_name, matched_node)
        #      # after insert dummy biasadd, that is Conv+dummybiasadd+add
        #      return self.apply_deconv2d_biasadd_fusion(new_match_node_name)

        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        _, q_inputs = self._get_node_input(normal_inputs[2])
        _, q_weights_inputs = self._get_node_input(normal_inputs[1])

        quantizev2_weights_name = q_weights_inputs[0]
        _, weights_name = self._get_node_input(quantizev2_weights_name)
        weights_min_name = weights_name[1]
        weights_max_name = weights_name[2]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weights_name[0]].node, self.per_channel
        )

        all_input_names = [normal_inputs[0]] + [q_weights_name] + q_inputs[:1]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        all_input_names += q_inputs[1:]
        skip_node_name.append(normal_inputs[2])
        skip_node_name.append(normal_inputs[1])
        skip_node_name.append(weights_name[0])
        skip_node_name.append(weights_min_name)
        skip_node_name.append(weights_max_name)
        skip_node_name.append(quantizev2_weights_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_deconv"

                bias_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                quantized_node_input_names = (
                    all_input_names[:3] + [bias_node_name] + all_input_names[3:] + control_inputs
                )

                node_op = "_FusedQuantizedDeconv2D"
                quantized_deconv_node = helper.create_node(node_op, quantized_node_name, quantized_node_input_names)

                helper.copy_attr(quantized_deconv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_deconv_node, "padding", node.attr["padding"])
                helper.copy_attr(quantized_deconv_node, "data_format", node.attr["data_format"])
                if "explicit_paddings" in node.attr:
                    helper.copy_attr(quantized_deconv_node, "explicit_paddings", node.attr["explicit_paddings"])
                helper.copy_attr(quantized_deconv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_deconv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_deconv_node, "Tfilter", dtypes.qint8)
                # helper.set_attr_string(quantized_conv_node, '_kernel', b'QuantizedMklOp')
                helper.set_attr_dtype(quantized_deconv_node, "out_type", dtypes.qint32)
                # helper.set_attr_dtype(quantized_conv_node, "alpha", dtypes.quint8)
                helper.set_attr_dtype(quantized_deconv_node, "Tbias", dtypes.float32)
                # if self.device == 'gpu' else dtypes.qint32)
                helper.set_attr_string_list(quantized_deconv_node, "fused_ops", [b"BiasAdd"])

                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_inputs",
                    [
                        dtypes.int32.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        input_data_type.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,  # if self.device == 'gpu' else dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_outputs",
                    [
                        dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )

                self.add_output_graph_node(quantized_deconv_node)
                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8, False)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, match_node_name[2], dtypes.qint8)

            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_single_deconv3d_fusion(self, match_node_name):
        """Apply single Conv3DBackpropInputV2 fusion.

        Dequantize + Conv3DBackpropInputV2 + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]

        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        _, q_inputs = self._get_node_input(normal_inputs[2])
        _, q_weights_inputs = self._get_node_input(normal_inputs[1])

        quantizev2_weights_name = q_weights_inputs[0]
        _, weights_name = self._get_node_input(quantizev2_weights_name)
        weights_min_name = weights_name[1]
        weights_max_name = weights_name[2]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weights_name[0]].node, self.per_channel
        )

        all_input_names = [normal_inputs[0]] + [q_weights_name] + q_inputs[:1]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        all_input_names += q_inputs[1:]
        skip_node_name.append(normal_inputs[2])
        skip_node_name.append(normal_inputs[1])
        skip_node_name.append(weights_name[0])
        skip_node_name.append(weights_min_name)
        skip_node_name.append(weights_max_name)
        skip_node_name.append(quantizev2_weights_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_deconv"

                quantized_node_input_names = all_input_names[:2] + all_input_names[2:] + control_inputs

                node_op = "_FusedQuantizedDeconv3D"
                quantized_deconv_node = helper.create_node(node_op, quantized_node_name, quantized_node_input_names)

                helper.copy_attr(quantized_deconv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_deconv_node, "padding", node.attr["padding"])
                helper.copy_attr(quantized_deconv_node, "data_format", node.attr["data_format"])
                if "explicit_paddings" in node.attr:
                    helper.copy_attr(quantized_deconv_node, "explicit_paddings", node.attr["explicit_paddings"])
                helper.copy_attr(quantized_deconv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_deconv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_deconv_node, "Tfilter", dtypes.qint8)
                # helper.set_attr_string(quantized_conv_node, '_kernel', b'QuantizedMklOp')
                helper.set_attr_dtype(quantized_deconv_node, "out_type", dtypes.qint32)
                # helper.set_attr_dtype(quantized_conv_node, "alpha", dtypes.quint8)

                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_inputs",
                    [
                        dtypes.int32.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        input_data_type.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_outputs",
                    [
                        dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )

                self.add_output_graph_node(quantized_deconv_node)
                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8, False)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, match_node_name[1], dtypes.qint8)

            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_deconv3d_biasadd_fusion(self, match_node_name):
        """Apply the Conv3DBackpropInputV2 BiasAdd fusion.

        Dequantize + Conv3DBackpropInputV2 + Biasadd + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]

        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        _, q_inputs = self._get_node_input(normal_inputs[2])
        _, q_weights_inputs = self._get_node_input(normal_inputs[1])

        quantizev2_weights_name = q_weights_inputs[0]
        _, weights_name = self._get_node_input(quantizev2_weights_name)
        weights_min_name = weights_name[1]
        weights_max_name = weights_name[2]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weights_name[0]].node, self.per_channel
        )

        all_input_names = [normal_inputs[0]] + [q_weights_name] + q_inputs[:1]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        all_input_names += q_inputs[1:]
        skip_node_name.append(normal_inputs[2])
        skip_node_name.append(normal_inputs[1])
        skip_node_name.append(weights_name[0])
        skip_node_name.append(weights_min_name)
        skip_node_name.append(weights_max_name)
        skip_node_name.append(quantizev2_weights_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_deconv"

                bias_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                quantized_node_input_names = (
                    all_input_names[:3] + [bias_node_name] + all_input_names[3:] + control_inputs
                )

                node_op = "_FusedQuantizedDeconv3D"
                quantized_deconv_node = helper.create_node(node_op, quantized_node_name, quantized_node_input_names)

                helper.copy_attr(quantized_deconv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_deconv_node, "padding", node.attr["padding"])
                helper.copy_attr(quantized_deconv_node, "data_format", node.attr["data_format"])
                if "explicit_paddings" in node.attr:
                    helper.copy_attr(quantized_deconv_node, "explicit_paddings", node.attr["explicit_paddings"])
                helper.copy_attr(quantized_deconv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_deconv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_deconv_node, "Tfilter", dtypes.qint8)
                # helper.set_attr_string(quantized_conv_node, '_kernel', b'QuantizedMklOp')
                helper.set_attr_dtype(quantized_deconv_node, "out_type", dtypes.qint32)
                # helper.set_attr_dtype(quantized_conv_node, "alpha", dtypes.quint8)
                helper.set_attr_dtype(quantized_deconv_node, "Tbias", dtypes.float32)
                # if self.device == 'gpu' else dtypes.qint32)
                helper.set_attr_string_list(quantized_deconv_node, "fused_ops", [b"BiasAdd"])

                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_inputs",
                    [
                        dtypes.int32.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        input_data_type.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,  # if self.device == 'gpu' else dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                helper.set_attr_type_list(
                    quantized_deconv_node,
                    "Thost_outputs",
                    [
                        dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )

                self.add_output_graph_node(quantized_deconv_node)
                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8, False)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, match_node_name[2], dtypes.qint8)

            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        """Get the longest fusion pattern."""
        self._get_op_list()

        matched_rule, matched_node_name = self._is_match_deconv(self.sorted_patterns)
        return matched_rule, matched_node_name

    def apply_the_transform(self):
        """Quantize Conv2DBackpropInput and Conv3DBackpropInputV2 and apply the fusion."""
        self._get_op_list()
        matched_rule, matched_node_name = self._is_match_deconv(self.sorted_patterns, True)
        if matched_node_name:
            self.output_graph = graph_pb2.GraphDef()
            fusion_name = "".join(matched_rule)
            if fusion_name in self.fusion_mapping:
                self.fusion_mapping[fusion_name](matched_node_name)
            else:  # pragma: no cover
                self.logger.info("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, self.exclude_deconv_nodes

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)

            return self.output_graph, self.exclude_deconv_nodes

        if self.remove_redundant_quant_flag:
            self.input_graph = self.remove_redundant_quantization(self.input_graph)
        return self.input_graph, self.exclude_deconv_nodes

    def _is_match_deconv(self, patterns, qdq_inserted=False):
        """Detect the rule matched nodes collections.

        Returns:
            [List] -- [the matched rule]
            [String] -- [the list contains the matched node name]
        """
        matched_node_name = []

        for k, v in enumerate(self.op_list):
            if v in set(fusion[1] for fusion in patterns):
                cur_node = self.node_name_mapping[list(self.node_name_mapping.keys())[k]].node
                if cur_node.name != self.start_node_name:
                    continue

                if qdq_inserted:
                    _, normal_inputs = self._get_node_input(cur_node.name)

                for sub_rule in patterns:
                    if sub_rule[0] != "Dequantize" or sub_rule[-1] != "QuantizeV2":
                        self.exclude_deconv_nodes.append(cur_node.name)
                        continue
                    if v != sub_rule[1]:
                        self.exclude_deconv_nodes.append(cur_node.name)
                        continue

                    if qdq_inserted:
                        input_index = 0
                        if sub_rule[1] in ("Conv2DBackpropInput", "Conv3DBackpropInputV2"):
                            input_index = 2
                        if (
                            self.node_name_mapping[normal_inputs[input_index]].node.op != "Dequantize"
                            or self.node_name_mapping[normal_inputs[1]].node.op != "Dequantize"
                        ):
                            continue

                    sub_rule_len = len(sub_rule) - 2
                    self.logger.debug("Try to apply rule: {}".format(sub_rule))

                    cur_node_name = list(self.node_name_mapping.keys())[k]

                    matched_node_name.clear()
                    matched_node_name.append(sub_rule[0])
                    matched_node_name.append(cur_node_name)

                    while sub_rule_len > 1:
                        if not self.node_name_mapping[cur_node_name].output:
                            self.logger.debug("Fail to match {}".format(sub_rule))
                            break

                        next_node_name = self.node_name_mapping[cur_node_name].output[0]

                        is_shared_output = True if len(self.node_name_mapping[cur_node_name].output) > 1 else False

                        next_node_op = self.node_name_mapping[next_node_name].node.op
                        if next_node_op == sub_rule[-sub_rule_len]:
                            if not is_shared_output:
                                matched_node_name.append(next_node_name)
                                sub_rule_len -= 1
                                cur_node_name = next_node_name
                            else:
                                matched_node_name.clear()
                                self.logger.debug("Fail to match {}.".format(sub_rule))
                                break
                        else:
                            matched_node_name.clear()
                            self.logger.debug("Fail to match {}.".format(sub_rule))
                            break

                    if sub_rule_len == 1:
                        matched_node_name.append(sub_rule[-1])
                        self.logger.debug("Match {} on nodes {}.".format(sub_rule, matched_node_name))
                        return sub_rule, matched_node_name

        return None, None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert fp32 op to int8 and fuse the pattern."""

from tensorflow.core.framework import graph_pb2
from tensorflow.python.platform import gfile

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from .quantize_graph_base import QuantizeGraphBase
from .quantize_graph_bn import FuseNodeStartWithFusedBatchNormV3
from .quantize_graph_concatv2 import FuseNodeStartWithConcatV2
from .quantize_graph_conv import FuseNodeStartWithConv2d
from .quantize_graph_matmul import FuseNodeStartWithMatmul
from .quantize_graph_pooling import FuseNodeStartWithPooling


class QuantizeGraphForIntel(QuantizeGraphBase):
    """Quantize the graph."""

    def __init__(
        self,
        input_graph,
        input_node_names,
        output_node_names,
        op_wise_config,
        op_wise_sequences,
        device,
        fake_quant=False,
        new_api=False,
        performance_only=False,
        itex_mode=False,
    ):
        """Quantize Graph For Intel Cpu."""
        super().__init__(output_node_names)
        self.op_wise_config = op_wise_config
        # self.perchannel = perchannel
        if isinstance(input_graph, graph_pb2.GraphDef):
            self.input_graph = input_graph
        else:
            self.input_graph = graph_pb2.GraphDef()
            with gfile.Open(input_graph, "rb") as f:
                self.input_graph.ParseFromString(f.read())

        input_output_names = input_node_names + output_node_names
        self.input_graph = QuantizeGraphHelper().remove_training_nodes(
            self.input_graph, protected_nodes=input_output_names
        )

        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.input_graph

        self.graph_info = self.graph_analyzer.parse_graph()
        self.graph_analyzer.get_frame_info()

        self.op_wise_seq = op_wise_sequences

        self.device = device
        self.fake_quant = fake_quant
        self.new_api = new_api
        self.performance_only = performance_only
        self.itex_mode = itex_mode

        self.all_quantizable_node = []
        self.exclude_node_names = []
        self.register_transformer("MaxPool", FuseNodeStartWithPooling)
        self.register_transformer("MaxPool3D", FuseNodeStartWithPooling)
        self.register_transformer("Conv2D", FuseNodeStartWithConv2d)
        self.register_transformer("Conv3D", FuseNodeStartWithConv2d)
        self.register_transformer("FusedBatchNormV3", FuseNodeStartWithFusedBatchNormV3)
        self.register_transformer("DepthwiseConv2dNative", FuseNodeStartWithConv2d)
        self.register_transformer("AvgPool", FuseNodeStartWithPooling)
        self.register_transformer("ConcatV2", FuseNodeStartWithConcatV2)
        self.register_transformer("MatMul", FuseNodeStartWithMatmul)

    @dump_elapsed_time("Pass Quantization")
    def do_transform(self):
        """Apply all the transformers to fuse into int8 op."""
        count = 0
        remove_redundant_quant_flag = False
        op_wise_config_name_list = list(self.op_wise_config.keys())
        all_node_length = len(self.op_wise_config)
        for _, node in enumerate(self.input_graph.node):
            if node in self.input_graph.node and node.op in self.transformers and node.name in self.op_wise_config:
                count += 1
                if count == all_node_length:
                    remove_redundant_quant_flag = True
                self.input_graph, quantizable_node_names, exclude_node_names = self.transformers[node.op](
                    input_graph=self.input_graph,
                    patterns=self.op_wise_seq[node.op],
                    remove_redundant_quant_flag=remove_redundant_quant_flag,
                    op_wise_cfg=self.op_wise_config[node.name],
                    op_wise_config_name_list=op_wise_config_name_list,
                    start_node_name=node.name,
                    device=self.device,
                    fake_quant=self.fake_quant,
                    new_api=self.new_api,
                    performance_only=self.performance_only,
                    itex_mode=self.itex_mode,
                    frame_info=self.graph_analyzer.parent_frame_details,
                ).apply_the_transform()
                if quantizable_node_names:
                    if node.op in ("ConcatV2", "MaxPool", "MaxPool3D", "AvgPool"):
                        self.all_quantizable_node.extend([[i] for i in quantizable_node_names])
                    else:
                        self.all_quantizable_node.append(quantizable_node_names)
                if exclude_node_names:
                    self.exclude_node_names.extend(exclude_node_names)

        return (
            self.remove_dead_nodes(self.input_graph, self.output_node_names),
            self.all_quantizable_node,
            self.exclude_node_names,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize the AvgPool/MaxPool."""

import tensorflow as tf
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper
from neural_compressor.tensorflow.utils import version1_eq_version2, version1_gt_version2, version1_lt_version2

from .quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithPooling(QuantizeNodeBase):
    """Quantize the AvgPool and MaxPool."""

    def _add_pool_function(self, original_node, quantized_op_node):
        """Set quantized pooling node attributes."""
        pooling_type = (
            dtypes.quint8
            if version1_lt_version2(tf.version.VERSION, "2.6.0") or self._find_relu_node(original_node)
            else dtypes.qint8
        )
        helper.set_attr_dtype(quantized_op_node, "T", pooling_type)
        helper.copy_attr(quantized_op_node, "ksize", original_node.attr["ksize"])
        helper.copy_attr(quantized_op_node, "strides", original_node.attr["strides"])
        helper.copy_attr(quantized_op_node, "padding", original_node.attr["padding"])

    def _apply_pool_quantization(self):
        """Quantize AvgPool/MaxPool."""
        for _, v in self.node_name_mapping.items():
            # Tensorflow 2.5.0 enabled the s8 input for pooling op.
            # If the tf version is lower than 2.5.0, we need to confirm the input
            # data type of pooling is unsigned.
            if v.node.name == self.start_node_name and (
                version1_gt_version2(tf.version.VERSION, "2.5.0")
                or version1_lt_version2(tf.version.VERSION, "2.6.0")
                and self._find_relu_node(v.node)
            ):
                self.eightbitize_single_input_tensor_node(v.node, self._add_pool_function)
                self.quantizable_node_names.append(v.node.name)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(v.node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        """Only pooling op itself, no fusion pattern."""
        return 1

    def apply_the_transform(self):
        """Quantize AvgPool/MaxPool."""
        self.quantizable_node_names = []
        self._apply_pool_quantization()
        self._reset_output_node_maps()
        if self.remove_redundant_quant_flag:
            self.output_graph = self.remove_redundant_quantization(self.output_graph)
        return self.output_graph, self.quantizable_node_names, []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse QuantizedMatMul with Requantize/Dequantize Graph Rewriter."""

import numpy as np
import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import version1_gt_version2, version1_lt_version2

from ..graph_base import GraphRewriterBase


class FuseMatMulRequantizeDequantizeTransformer(GraphRewriterBase):
    """Fuse QuantizedMatMul + Requantize + Dequantize into QuantizedMatMulWithBiasAndDequantize."""

    def __init__(self, model, device="cpu"):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

        self.eps = 1e-5

    def do_transformation(self):
        """Apply the fusion of QuantizedMatMul + Requantize + Dequantize."""
        fuse_pattern = []
        if tf.version.VERSION in ("1.15.0-up2", "1.15.0-up3") or version1_gt_version2(tf.version.VERSION, "2.1.0"):
            fuse_pattern = [["QuantizedMatMulWithBias"], ["Requantize"], ["Dequantize"], ("Softmax",)]
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(fuse_pattern)
        for i in target_nodes:
            # TODO Remove below checker once the TF's limitation removed.
            if len(i) == 5 and version1_lt_version2(tf.__version__, "2.6.0"):
                continue

            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]
            deq_node_name = i[2]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndDequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if "T1" in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr["T1"])
            if "T2" in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr["T2"])

            top_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = self.graph_info[new_node.input[6]].node
            min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node

            weight_node = self.graph_info[Helper.node_name_from_input(new_node.input[1])].node
            bias_node = self.graph_info[Helper.node_name_from_input(new_node.input[2])].node
            max_input_node = self.graph_info[last_node.input[-1]].node
            min_input_node = self.graph_info[last_node.input[-2]].node

            if max_input_node.op == "Enter":  # pragma: no cover
                min_input_parent_name = Helper.node_name_from_input(min_input_node.input[0])
                max_input_parent_name = Helper.node_name_from_input(max_input_node.input[0])
                min_input_parent_node = self.graph_info[min_input_parent_name].node
                max_input_parent_node = self.graph_info[max_input_parent_name].node
                if min_input_parent_node.op != "Const" or max_input_parent_node.op != "Const":
                    continue
                min_input_node = min_input_parent_node
                max_input_node = max_input_parent_node
            if max_filter_node.op == "Enter":  # pragma: no cover
                min_filter_parent_name = Helper.node_name_from_input(min_filter_node.input[0])
                max_filter_parent_name = Helper.node_name_from_input(max_filter_node.input[0])
                min_filter_parent_node = self.graph_info[min_filter_parent_name].node
                max_filter_parent_node = self.graph_info[max_filter_parent_name].node
                if min_filter_parent_node.op != "Const" or max_filter_parent_node.op != "Const":
                    continue
                min_filter_node = min_filter_parent_node
                max_filter_node = max_filter_parent_node
            if weight_node.op == "Enter":  # pragma: no cover
                weight_parent_name = Helper.node_name_from_input(weight_node.input[0])
                weight_parent_node = self.graph_info[weight_parent_name].node
                if weight_parent_node.op != "Const":
                    continue
                weight_node = weight_parent_node
            bias_enter_node = None
            if bias_node.op == "Enter":  # pragma: no cover
                bias_enter_node = bias_node
                bias_parent_name = Helper.node_name_from_input(bias_node.input[0])
                bias_parent_node = self.graph_info[bias_parent_name].node
                if bias_parent_node.op != "Const":
                    continue
                bias_node = bias_parent_node

            if max_filter_node.op == "Const":
                min_input_value = (min_input_node.attr["value"].tensor.float_val)[0]
                max_input_value = (max_input_node.attr["value"].tensor.float_val)[0]

                max_filter_value = (max_filter_node.attr["value"].tensor.float_val)[0]
                min_filter_value = (min_filter_node.attr["value"].tensor.float_val)[0]

                weights_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                bias_tensor = tensor_util.MakeNdarray(bias_node.attr["value"].tensor)
                is_min_first = bool(quantized_node.attr["input_quant_mode"].s == b"MIN_FIRST")
                input_range = (
                    max_input_value - min_input_value
                    if is_min_first
                    else max(abs(max_input_value), abs(min_input_value))
                )

                if -self.eps <= input_range <= self.eps:
                    input_range += self.eps

                if -self.eps <= max_input_value - min_input_value <= self.eps:
                    max_input_value += self.eps
                int32_bias = Helper.generate_int32_bias_for_matmul(
                    bias_tensor,
                    weights_tensor,
                    input_range,
                    max_input_value,
                    min_input_value,
                    max_filter_value,
                    min_filter_value,
                )

                bias_node.attr["dtype"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )
                bias_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            bias_tensor if self.device == "gpu" else int32_bias,
                            dtypes.float32 if self.device == "gpu" else dtypes.int32,
                            bias_tensor.shape,
                        )
                    )
                )

                bias_node.attr["value"].tensor.dtype = float32_type if self.device == "gpu" else qint32_type
                new_node.attr["Tbias"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )
                if bias_enter_node:
                    bias_enter_node.attr["T"].CopyFrom(
                        attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                    )
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            new_node.attr["Toutput"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            self.graph_analyzer.remove_node(requantize_node_name)

            if self.graph_info[deq_node_name].outputs:
                self.graph_analyzer.replace_single_node(
                    new_node,
                    [top_node_name],
                    quantized_node_name,
                    self.graph_info[deq_node_name].outputs,
                    deq_node_name,
                )
                self.graph_analyzer.remove_node(deq_node_name)
            else:
                self.graph_analyzer.remove_node(deq_node_name)

                new_node.name = deq_node_name
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name, [], deq_node_name
                )

            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()


class FuseMatMulRequantizeTransformer(GraphRewriterBase):
    """Fuse Quantized MatMul Op with the successor Requantize Op."""

    def __init__(self, model, device="cpu"):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.eps = 1e-05
        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.

        Returns:
            [graphdef]: the optimized graphdef object
        """
        uint8_type = dtypes.quint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum

        while True:
            target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(
                [["QuantizedMatMulWithBiasAndRelu"], ["Requantize"]]
            )
            if len(target_nodes) == 0:
                break

            i = target_nodes[0]
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)
            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if "T1" in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr["T1"])
            if "T2" in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr["T2"])

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = self.graph_info[new_node.input[6]].node
            min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node

            is_min_first = bool(quantized_node.attr["input_quant_mode"].s == b"MIN_FIRST")

            weight_node = self.graph_info[new_node.input[1]].node
            bias_node = self.graph_info[new_node.input[2]].node

            max_input_node = None
            min_input_node = None
            if last_node.op.find("Requantize") != -1 or last_node.op.find("QuantizeV2") != -1:
                max_input_node = self.graph_info[last_node.input[-1]].node
                min_input_node = self.graph_info[last_node.input[-2]].node

            if max_input_node and max_input_node.op == "Enter":  # pragma: no cover
                min_input_parent_name = Helper.node_name_from_input(min_input_node.input[0])
                max_input_parent_name = Helper.node_name_from_input(max_input_node.input[0])
                min_input_parent_node = self.graph_info[min_input_parent_name].node
                max_input_parent_node = self.graph_info[max_input_parent_name].node
                if min_input_parent_node.op != "Const" or max_input_parent_node.op != "Const":
                    continue
                min_input_node = min_input_parent_node
                max_input_node = max_input_parent_node
            if max_filter_node.op == "Enter":  # pragma: no cover
                min_filter_parent_name = Helper.node_name_from_input(min_filter_node.input[0])
                max_filter_parent_name = Helper.node_name_from_input(max_filter_node.input[0])
                min_filter_parent_node = self.graph_info[min_filter_parent_name].node
                max_filter_parent_node = self.graph_info[max_filter_parent_name].node
                if min_filter_parent_node.op != "Const" or max_filter_parent_node.op != "Const":
                    continue
                min_filter_node = min_filter_parent_node
                max_filter_node = max_filter_parent_node
            if weight_node.op == "Enter":  # pragma: no cover
                weight_parent_name = Helper.node_name_from_input(weight_node.input[0])
                weight_parent_node = self.graph_info[weight_parent_name].node
                if weight_parent_node.op != "Const":
                    continue
                weight_node = weight_parent_node
            bias_enter_node = None
            if bias_node.op == "Enter":  # pragma: no cover
                bias_enter_node = bias_node
                bias_parent_name = Helper.node_name_from_input(bias_node.input[0])
                bias_parent_node = self.graph_info[bias_parent_name].node
                if bias_parent_node.op != "Const":
                    continue
                bias_node = bias_parent_node

            if last_node.op.find("Requantize") != -1 or last_node.op.find("QuantizeV2") != -1:
                min_input_value = (min_input_node.attr["value"].tensor.float_val)[0]
                max_input_value = (max_input_node.attr["value"].tensor.float_val)[0]

                max_filter_value = (max_filter_node.attr["value"].tensor.float_val)[0]
                min_filter_value = (min_filter_node.attr["value"].tensor.float_val)[0]

                weights_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                bias_tensor = tensor_util.MakeNdarray(bias_node.attr["value"].tensor)

                input_range = (
                    max_input_value - min_input_value
                    if is_min_first
                    else max(abs(max_input_value), abs(min_input_value))
                )

                if -self.eps <= input_range <= self.eps:
                    input_range += self.eps

                if -self.eps <= max_input_value - min_input_value <= self.eps:
                    max_input_value += self.eps
                int32_bias = Helper.generate_int32_bias_for_matmul(
                    bias_tensor,
                    weights_tensor,
                    input_range,
                    max_input_value,
                    min_input_value,
                    max_filter_value,
                    min_filter_value,
                )
                bias_node.attr["dtype"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )
                bias_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            bias_tensor if self.device == "gpu" else int32_bias,
                            dtypes.float32 if self.device == "gpu" else dtypes.int32,
                            bias_tensor.shape,
                        )
                    )
                )

                bias_node.attr["value"].tensor.dtype = float32_type if self.device == "gpu" else qint32_type
                new_node.attr["Tbias"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )

                new_node.attr["Toutput"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
                # TODO enabled below commit once the graph refactor pre_optimize committed.
                if quantized_node_op.find("Relu") == -1:
                    deq_node_name = self.graph_info[requantize_node_name].outputs[0]
                    deq_node = self.graph_info[deq_node_name].node
                    deq_node.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
                if bias_enter_node:
                    bias_enter_node.attr["T"].CopyFrom(
                        attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                    )
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            self.graph_analyzer.replace_single_node(
                new_node,
                [parent_node_name],
                quantized_node_name,
                [self.graph_info[requantize_node_name].outputs[0]],
                requantize_node_name,
            )
            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()


class FuseMatMulRequantizeDequantizeNewAPITransformer(GraphRewriterBase):  # pragma: no cover
    """Fuse _QuantizedMatMul + Requantize + Dequantize into _QuantizedMatMul."""

    def __init__(self, model, device="cpu"):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

        self.eps = 1e-5

    def do_transformation(self):
        """Apply the fusion of QuantizedMatMul + Requantize + Dequantize."""
        fuse_pattern = [["_QuantizedMatMul"], ["Requantize", "RequantizePerChannel"], ["Dequantize"], ("Softmax",)]

        uint8_type = dtypes.quint8.as_datatype_enum
        int8_type = dtypes.qint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(fuse_pattern)
        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]
            deq_node_name = i[2]

            quantized_node_op = i[-1][0]

            # "BiasAdd" + "Add" only supports "Dequantize"
            attr_fused_ops = "".join(
                x
                for x in quantized_node.attr["fused_ops"].SerializeToString().decode("UTF-8", "ignore").strip()
                if x.isprintable()
            )
            if "BiasAddAdd" not in attr_fused_ops:
                continue

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if "T1" in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr["T1"])
            if "T2" in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr["T2"])
            if "U" in quantized_node.attr:
                new_node.attr["U"].CopyFrom(quantized_node.attr["U"])
            if "transpose_b" in quantized_node.attr:
                new_node.attr["transpose_b"].CopyFrom(quantized_node.attr["transpose_b"])
            if "transpose_a" in quantized_node.attr:
                new_node.attr["transpose_a"].CopyFrom(quantized_node.attr["transpose_a"])
            if "input_quant_mode" in quantized_node.attr:
                new_node.attr["input_quant_mode"].CopyFrom(quantized_node.attr["input_quant_mode"])
            if "output_quant_mode" in quantized_node.attr:
                new_node.attr["output_quant_mode"].CopyFrom(quantized_node.attr["output_quant_mode"])

            top_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = None
            min_filter_node = None

            # MatMul + BiasAdd + Add
            # The Min and Max of non-const weight node are from QuantizeV2's output, not valid nodes.
            # Add check here for excluding this case.
            if ":2" not in new_node.input[7]:
                max_filter_node = self.graph_info[new_node.input[7]].node
            if ":1" not in new_node.input[6]:
                min_filter_node = self.graph_info[new_node.input[6]].node
            last_node = self.graph_info[new_node.input[0]].node
            weight_node = self.graph_info[Helper.node_name_from_input(new_node.input[1])].node
            bias_node = self.graph_info[Helper.node_name_from_input(new_node.input[2])].node
            if not last_node.op == "QuantizedConcatV2":
                max_input_node = self.graph_info[last_node.input[-1]].node
                min_input_node = self.graph_info[last_node.input[-2]].node

            type_bias = float32_type
            if not last_node.op == "QuantizedConcatV2" and max_input_node.op == "Enter":  # pragma: no cover
                min_input_parent_name = Helper.node_name_from_input(min_input_node.input[0])
                max_input_parent_name = Helper.node_name_from_input(max_input_node.input[0])
                min_input_parent_node = self.graph_info[min_input_parent_name].node
                max_input_parent_node = self.graph_info[max_input_parent_name].node
                if min_input_parent_node.op != "Const" or max_input_parent_node.op != "Const":
                    continue
                min_input_node = min_input_parent_node
                max_input_node = max_input_parent_node
            if max_filter_node and min_filter_node and max_filter_node.op == "Enter":  # pragma: no cover
                min_filter_parent_name = Helper.node_name_from_input(min_filter_node.input[0])
                max_filter_parent_name = Helper.node_name_from_input(max_filter_node.input[0])
                min_filter_parent_node = self.graph_info[min_filter_parent_name].node
                max_filter_parent_node = self.graph_info[max_filter_parent_name].node
                if min_filter_parent_node.op != "Const" or max_filter_parent_node.op != "Const":
                    continue
                min_filter_node = min_filter_parent_node
                max_filter_node = max_filter_parent_node
            if weight_node.op == "Enter":  # pragma: no cover
                weight_parent_name = Helper.node_name_from_input(weight_node.input[0])
                weight_parent_node = self.graph_info[weight_parent_name].node
                if weight_parent_node.op != "Const":
                    continue
                weight_node = weight_parent_node
            bias_enter_node = None
            if bias_node.op == "Enter":  # pragma: no cover
                bias_enter_node = bias_node
                bias_parent_name = Helper.node_name_from_input(bias_node.input[0])
                bias_parent_node = self.graph_info[bias_parent_name].node
                if bias_parent_node.op != "Const":
                    continue
                bias_node = bias_parent_node

            if (
                max_filter_node
                and min_filter_node
                and max_filter_node.op == "Const"
                and weight_node.op == "Const"
                and not last_node.op == "QuantizedConcatV2"
            ):
                min_input_value = (min_input_node.attr["value"].tensor.float_val)[0]
                max_input_value = (max_input_node.attr["value"].tensor.float_val)[0]
                if requantize_node.op.find("PerChannel") != -1:  # pragma: no cover
                    max_filter_tensor = tensor_util.MakeNdarray(max_filter_node.attr["value"].tensor)  # get tensor
                    min_filter_tensor = tensor_util.MakeNdarray(min_filter_node.attr["value"].tensor)  # get tensor
                else:
                    max_filter_value = (max_filter_node.attr["value"].tensor.float_val)[0]
                    min_filter_value = (min_filter_node.attr["value"].tensor.float_val)[0]

                weights_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                bias_tensor = tensor_util.MakeNdarray(bias_node.attr["value"].tensor)
                is_min_first = bool(quantized_node.attr["input_quant_mode"].s == b"MIN_FIRST")
                input_range = (
                    max_input_value - min_input_value
                    if is_min_first
                    else max(abs(max_input_value), abs(min_input_value))
                )

                if -self.eps <= input_range <= self.eps:
                    input_range += self.eps

                if -self.eps <= max_input_value - min_input_value <= self.eps:
                    max_input_value += self.eps
                if requantize_node.op.find("PerChannel") != -1:  # pragma: no cover
                    int32_bias = Helper.generate_int32_bias_for_matmul_per_channel(
                        bias_tensor,
                        weights_tensor,
                        max_input_value,
                        min_input_value,
                        max_filter_tensor,
                        min_filter_tensor,
                    )
                else:
                    int32_bias = Helper.generate_int32_bias_for_matmul(
                        bias_tensor,
                        weights_tensor,
                        input_range,
                        max_input_value,
                        min_input_value,
                        max_filter_value,
                        min_filter_value,
                    )

                bias_node.attr["dtype"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )
                bias_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            bias_tensor if self.device == "gpu" else int32_bias,
                            dtypes.float32 if self.device == "gpu" else dtypes.int32,
                            bias_tensor.shape,
                        )
                    )
                )

                bias_node.attr["value"].tensor.dtype = float32_type if self.device == "gpu" else qint32_type
                type_bias = float32_type if self.device == "gpu" else qint32_type
                if bias_enter_node:
                    bias_enter_node.attr["T"].CopyFrom(
                        attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                    )
            else:
                type_bias = float32_type

            new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=type_bias))

            Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Add", b"Dequantize"])
            Helper.set_attr_type_list(
                new_node,
                "Thost_inputs",
                [
                    uint8_type,
                    int8_type,
                    type_bias,
                    float32_type,
                    float32_type,
                    float32_type,
                    float32_type,
                    float32_type,
                    float32_type,
                    float32_type,
                ],
            )

            Helper.set_attr_type_list(new_node, "Thost_outputs", [float32_type])
            new_node.attr["Tout"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            self.graph_analyzer.remove_node(requantize_node_name)

            if self.graph_info[deq_node_name].outputs:
                self.graph_analyzer.replace_single_node(
                    new_node,
                    [top_node_name],
                    quantized_node_name,
                    self.graph_info[deq_node_name].outputs,
                    deq_node_name,
                )
                self.graph_analyzer.remove_node(deq_node_name)
            else:
                self.graph_analyzer.remove_node(deq_node_name)

                new_node.name = deq_node_name
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name, [], deq_node_name
                )

            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()


class FuseMatMulRequantizeNewAPITransformer(GraphRewriterBase):
    """Fuse newAPI Quantized MatMul Op with the successor Requantize Op."""

    def __init__(self, model, device="cpu"):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.eps = 1e-05
        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.

        Returns:
            [graphdef]: the optimized graphdef object
        """
        uint8_type = dtypes.quint8.as_datatype_enum
        int8_type = dtypes.qint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum

        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(
            [["_QuantizedMatMul"], ["Requantize", "RequantizePerChannel"]]
        )
        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]
            attr_fused_ops = "".join(
                x
                for x in quantized_node.attr["fused_ops"].SerializeToString().decode("UTF-8", "ignore").strip()
                if x.isprintable()
            )
            # "Requantize"
            # "BiasAdd", "Requantize"
            # "BiasAdd", "Activation", "Requantize"
            if "BiasAddAdd" in attr_fused_ops:
                continue
            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)
            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)

            if "transpose_b" in quantized_node.attr:
                new_node.attr["transpose_b"].CopyFrom(quantized_node.attr["transpose_b"])
            if "transpose_a" in quantized_node.attr:
                new_node.attr["transpose_a"].CopyFrom(quantized_node.attr["transpose_a"])
            if "T1" in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr["T1"])
            if "T2" in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr["T2"])
            if "U" in quantized_node.attr:
                new_node.attr["U"].CopyFrom(quantized_node.attr["U"])
            if "input_quant_mode" in quantized_node.attr:
                new_node.attr["input_quant_mode"].CopyFrom(quantized_node.attr["input_quant_mode"])
            if "output_quant_mode" in quantized_node.attr:
                new_node.attr["output_quant_mode"].CopyFrom(quantized_node.attr["output_quant_mode"])

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = None
            min_filter_node = None
            # The Min and Max of non-const weight node are from QuantizeV2's output, not valid nodes.
            # Add check here for excluding this case.
            if len(attr_fused_ops) == 0:  # single matmul case
                if ":2" not in new_node.input[5]:
                    max_filter_node = self.graph_info[new_node.input[5]].node
                if ":1" not in new_node.input[4]:
                    min_filter_node = self.graph_info[new_node.input[4]].node
            else:
                if ":2" not in new_node.input[6]:
                    max_filter_node = self.graph_info[new_node.input[6]].node
                if ":1" not in new_node.input[5]:
                    min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node
            is_min_first = bool(quantized_node.attr["input_quant_mode"].s == b"MIN_FIRST")
            weight_node = self.graph_info[new_node.input[1]].node
            bias_node = None
            if "BiasAdd" in attr_fused_ops:
                bias_node = self.graph_info[new_node.input[2]].node
            if not last_node.op == "QuantizedConcatV2":
                max_input_node = self.graph_info[last_node.input[-1]].node
                min_input_node = self.graph_info[last_node.input[-2]].node

            if not last_node.op == "QuantizedConcatV2" and max_input_node.op == "Enter":  # pragma: no cover
                min_input_parent_name = Helper.node_name_from_input(min_input_node.input[0])
                max_input_parent_name = Helper.node_name_from_input(max_input_node.input[0])
                min_input_parent_node = self.graph_info[min_input_parent_name].node
                max_input_parent_node = self.graph_info[max_input_parent_name].node
                if min_input_parent_node.op != "Const" or max_input_parent_node.op != "Const":
                    continue
                min_input_node = min_input_parent_node
                max_input_node = max_input_parent_node
            if (
                max_filter_node
                and min_filter_node
                and min_filter_node.input
                and max_filter_node.input
                and max_filter_node.op == "Enter"
            ):  # pragma: no cover
                min_filter_parent_name = Helper.node_name_from_input(min_filter_node.input[0])
                max_filter_parent_name = Helper.node_name_from_input(max_filter_node.input[0])
                min_filter_parent_node = self.graph_info[min_filter_parent_name].node
                max_filter_parent_node = self.graph_info[max_filter_parent_name].node
                if min_filter_parent_node.op != "Const" or max_filter_parent_node.op != "Const":
                    continue
                min_filter_node = min_filter_parent_node
                max_filter_node = max_filter_parent_node
            if weight_node.op == "Enter":  # pragma: no cover
                weight_parent_name = Helper.node_name_from_input(weight_node.input[0])
                weight_parent_node = self.graph_info[weight_parent_name].node
                if weight_parent_node.op != "Const":
                    continue
                weight_node = weight_parent_node
            bias_enter_node = None
            if bias_node and bias_node.op == "Enter":  # pragma: no cover
                bias_enter_node = bias_node
                bias_parent_name = Helper.node_name_from_input(bias_node.input[0])
                bias_parent_node = self.graph_info[bias_parent_name].node
                if bias_parent_node.op != "Const":
                    continue
                bias_node = bias_parent_node

            if bias_node and (
                last_node.op.find("_QuantizedMatMul") != -1
                or last_node.op.find("QuantizeV2") != -1
                and max_filter_node
                and min_filter_node
            ):
                min_input_value = (min_input_node.attr["value"].tensor.float_val)[0]
                max_input_value = (max_input_node.attr["value"].tensor.float_val)[0]
                if requantize_node.op.find("PerChannel") != -1:  # pragma: no cover
                    max_filter_tensor = tensor_util.MakeNdarray(max_filter_node.attr["value"].tensor)  # get tensor
                    min_filter_tensor = tensor_util.MakeNdarray(min_filter_node.attr["value"].tensor)  # get tensor
                else:
                    max_filter_value = (max_filter_node.attr["value"].tensor.float_val)[0]
                    min_filter_value = (min_filter_node.attr["value"].tensor.float_val)[0]

                weights_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                bias_tensor = tensor_util.MakeNdarray(bias_node.attr["value"].tensor)

                input_range = (
                    max_input_value - min_input_value
                    if is_min_first
                    else max(abs(max_input_value), abs(min_input_value))
                )
                if -self.eps <= input_range <= self.eps:
                    input_range += self.eps

                if -self.eps <= max_input_value - min_input_value <= self.eps:
                    max_input_value += self.eps
                if requantize_node.op.find("PerChannel") != -1:  # pragma: no cover
                    int32_bias = Helper.generate_int32_bias_for_matmul_per_channel(
                        bias_tensor,
                        weights_tensor,
                        max_input_value,
                        min_input_value,
                        max_filter_tensor,
                        min_filter_tensor,
                    )
                else:
                    int32_bias = Helper.generate_int32_bias_for_matmul(
                        bias_tensor,
                        weights_tensor,
                        input_range,
                        max_input_value,
                        min_input_value,
                        max_filter_value,
                        min_filter_value,
                    )
                bias_node.attr["dtype"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )
                bias_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            bias_tensor if self.device == "gpu" else int32_bias,
                            dtypes.float32 if self.device == "gpu" else dtypes.int32,
                            bias_tensor.shape,
                        )
                    )
                )

                bias_node.attr["value"].tensor.dtype = float32_type if self.device == "gpu" else qint32_type
                if bias_enter_node:
                    bias_enter_node.attr["T"].CopyFrom(
                        attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                    )
                new_node.attr["Tbias"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )

                deq_node_name = self.graph_info[requantize_node_name].outputs[0]
                deq_node = self.graph_info[deq_node_name].node
                deq_node.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

                Helper.set_attr_type_list(
                    new_node,
                    "Thost_inputs",
                    [
                        uint8_type,
                        int8_type,
                        float32_type if self.device == "gpu" else qint32_type,
                        float32_type,
                        float32_type,
                        float32_type,
                        float32_type,
                        float32_type,
                        float32_type,
                    ],
                )
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
                deq_node_name = self.graph_info[requantize_node_name].outputs[0]
                deq_node = self.graph_info[deq_node_name].node
                deq_node.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
                if bias_node:
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            uint8_type,
                            int8_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                        ],
                    )
                else:
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            uint8_type,
                            int8_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                            float32_type,
                        ],
                    )

            Helper.set_attr_type_list(new_node, "Thost_outputs", [uint8_type, float32_type, float32_type])

            if "GeluApproximate" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"GeluApproximate", b"Requantize"])
            elif "GeluExact" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"GeluExact", b"Requantize"])
            elif "Elu" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Elu", b"Requantize"])
            elif "LeakyRelu" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"LeakyRelu", b"Requantize"])
            elif "Relu6" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Relu6", b"Requantize"])
            elif "Tanh" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Tanh", b"Requantize"])
            elif "Sigmoid" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Sigmoid", b"Requantize"])
            elif "Relu" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Relu", b"Requantize"])
            elif "BiasAdd" in attr_fused_ops:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Requantize"])
            else:
                Helper.set_attr_string_list(new_node, "fused_ops", [b"Requantize"])
            new_node.attr["Tout"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

            self.graph_analyzer.replace_single_node(
                new_node,
                [parent_node_name],
                quantized_node_name,
                [self.graph_info[requantize_node_name].outputs[0]],
                requantize_node_name,
            )
            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse BiasAdd and Add Graph Rewriter."""

import tensorflow as tf
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from ..graph_base import GraphRewriterBase


class FuseBiasAddAndAddOptimizer(GraphRewriterBase):
    """Fuse Biasadd + Add into BiasAdd when the second input of Add is const node."""

    def do_transformation(self):
        """Fuse Biasadd + Add into BiasAdd for pattern fusion."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()

        target_nodes = cur_graph.query_fusion_pattern_nodes(
            [["Conv2D", "Conv3D"], "BiasAdd", ["Add", "AddV2"], ["Relu", "Relu6", "swish_f32"], ["Mul"], ["Mul"]]
        )

        for i in target_nodes:
            biasadd_const_name = graph_info[i[1]].node.input[1]
            biasadd_const_node = graph_info[biasadd_const_name].node

            if len(graph_info[i[1]].outputs) > 1:
                continue

            another_node_index = None
            for index, value in enumerate(graph_info[i[2]].node.input):
                if value != i[1]:
                    another_node_index = index
                    break
            add_node_const_name = graph_info[i[2]].node.input[another_node_index]

            add_const_node = graph_info[add_node_const_name].node

            if add_const_node.op != "Const":
                continue
            value = tensor_util.MakeNdarray(biasadd_const_node.attr["value"].tensor)
            add_value = tensor_util.MakeNdarray(add_const_node.attr["value"].tensor)

            new_bias_tensor = value + add_value
            fused_const_node = Helper.create_constant_node(i[2] + "_fused", new_bias_tensor, dtypes.float32)
            cur_graph.remove_node(graph_info[i[1]].node.input[1])

            graph_info[i[1]].node.input[1] = i[2] + "_fused"

            cur_graph.remove_node(add_node_const_name)

            cur_graph.remove_node(i[2])
            graph_info[i[3]].node.input[0] = i[1]
            cur_graph.add_node(fused_const_node, None, [i[1]])

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fetch Weight from Reshape Graph Rewriter."""


import numpy as np
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class FetchWeightFromReshapeOptimizer(GraphRewriterBase):
    """Handle the Pack + Reshape + Conv2D fusion pattern."""

    @dump_elapsed_time("Pass FetchWeightFromReshapeOptimizer")
    def do_transformation(self):
        """Fetch weight of Conv2D from Pack+Reshape+Conv2D pattern.

        Args:
          input_graph_def (graphdef): graphdef object
        Returns:
           [graphdef]: optimized graph
        """
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        target_nodes = cur_graph.query_fusion_pattern_nodes([["Pack"], ["Reshape"], ["Conv2D"]])

        for i, node_combination in enumerate(target_nodes):
            pack_node = graph_info[node_combination[0]].node
            reshape_node = graph_info[node_combination[1]].node
            shape_node = graph_info[reshape_node.input[1]].node
            conv_node = graph_info[node_combination[2]].node
            if not (pack_node.op == "Pack" and reshape_node.op == "Reshape" and conv_node.op == "Conv2D"):
                continue
            reshape_outputs_length = len(graph_info[node_combination[1]].outputs)
            unpack_values = []
            for index in range(pack_node.attr["N"].i):
                values_node = graph_info[pack_node.input[index]].node
                if values_node.op == "Const":
                    unpack_values.append(Helper.values_from_const(values_node))
            input_reshape = np.stack(unpack_values, axis=pack_node.attr["axis"].i)
            if shape_node.op != "Const":
                continue
            shape = Helper.values_from_const(shape_node)
            weight = np.reshape(input_reshape, shape)
            weight_node = Helper.create_constant_node(
                reshape_node.name + "/weight" + "_" + str(i), weight, dtypes.float32
            )
            if i > 0:
                conv_node_j = graph_info[target_nodes[i - 1][2]].node
                graph_info[node_combination[1]].outputs.remove(conv_node_j.name)
            for output in graph_info[node_combination[1]].outputs:
                successor_node = graph_info[output].node
                replace_index = None
                for index, value in enumerate(successor_node.input):
                    if value == reshape_node.name or value == reshape_node.name + "/weight" + "_" + str(i - 1):
                        replace_index = index
                        break
                # weight->conv2d
                cur_graph.add_node(weight_node, None, [successor_node.name])
                successor_node.input[replace_index] = weight_node.name

            if i + 1 == reshape_outputs_length:
                cur_graph.remove_node(reshape_node.name)
                cur_graph.remove_node(values_node.name)
                cur_graph.remove_node(shape_node.name)
                cur_graph.remove_node(pack_node.name)

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Insert QDQ pattern Graph Rewriter."""

import copy
import re
from collections import namedtuple

import numpy as np
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class GenerateGraphWithQDQPattern(GraphRewriterBase):
    """Insert Q/DQ pairs before quantizable ops."""

    def __init__(
        self,
        model,
        calibration_data,
        op_wise_config,
        fake_quant,
        fp32_ops,
        bf16_ops,
        quantized_nodes,
        device,
        performance_only,
        itex_mode,
        llm_weight_minmax,
    ):
        """Initialization."""
        super().__init__(model)
        self.data = calibration_data
        self.op_wise_config = op_wise_config
        self.fake_quant = fake_quant
        self.fp32_ops = fp32_ops
        self.bf16_ops = bf16_ops
        self.quantized_nodes = quantized_nodes
        self.device = device
        self.performance_only = performance_only
        self.itex_mode = itex_mode
        self.llm_weight_minmax = llm_weight_minmax
        self.node_details = namedtuple("node_details", ["node", "output"])
        self.node_name_mapping = {}
        self.check_op_list = {
            "ConcatV2",
            "Conv2D",
            "Conv3D",
            "DepthwiseConv2D",
            "QuantizeV2",
            "DepthwiseConv2dNative",
            "MaxPool",
            "MaxPool3D",
            "FusedBatchNormV3",
            "Requantize",
            "RequantizePerChannel",
            "AvgPool",
            "Pad",
            "CropAndResize",
            "Dequantize",
            "Mean",
            "MatMul",
            "BatchMatMul",
            "BatchMatMulV2",
            "FakeQuantWithMinMaxVars",
            "_MklFusedInstanceNorm",
            "Conv2DBackpropInput",
            "Conv3DBackpropInputV2",
            "Sigmoid",
        }

        for node in self.model.node:
            if node.name in self.node_name_mapping:
                raise ValueError("Duplicate Node Found when _parse_graph, the node name is {}".format(node.name))
            self.node_name_mapping[node.name] = self.node_details(node=node, output=[])
        for node_name in self.node_name_mapping:
            for each_input in self.node_name_mapping[node_name].node.input:
                self.node_name_mapping[Helper.node_name_from_input(each_input)].output.append(node_name)

    @dump_elapsed_time("Pass GenerateGraphWithQDQPattern")
    def do_transformation(self):
        """Generate the graph with QDQ patterns, this is the first step to do new api quantizaiton."""
        min_max_values = {}
        for i in self.data:
            if i.find("_requant") == -1:
                key, value = i.rsplit(":", 1)[0], i.rsplit(":", 1)[1]
                key = key.split("_eightbit_")[0][1:] + key[-5:]
                if key not in min_max_values:
                    min_max_values[key] = [float(value[1:-1])]
                else:
                    min_max_values[key].append(float(value[1:-1]))
        quantizable_op_names = []
        for i in min_max_values:
            if i.split("__")[0] not in quantizable_op_names:
                quantizable_op_names.append(i.split("__")[0])

        self.g = GraphAnalyzer()
        self.g.graph = copy.deepcopy(self.model)
        self.graph_info = self.g.parse_graph()

        self.g.get_frame_info()

        # insert QDQ pattern for op's input
        for op_name in quantizable_op_names:
            if self._ignore_insert_qdq_pattern(op_name):
                continue

            if op_name not in self.op_wise_config.keys():
                is_asymmetric = False
            else:
                op_wise_cfg = self.op_wise_config[op_name]
                is_asymmetric = op_wise_cfg[2]
            if self.graph_info[op_name].node.op == "ConcatV2":
                if not self.itex_mode:
                    self._insert_qdq_pattern_for_concatv2(self.graph_info[op_name].node, is_asymmetric)
            else:
                self._insert_qdq_pattern_for_common_ops(self.graph_info[op_name].node, is_asymmetric)

        # insert QDQ pattern for op's weight
        self.g_weight = GraphAnalyzer()
        self.g_weight.graph = self.g.dump_graph()
        self.graph_info = self.g_weight.parse_graph()
        target_nodes = self.g_weight.query_fusion_pattern_nodes(
            [
                [
                    "Conv2D",
                    "Conv3D",
                    "DepthwiseConv2dNative",
                    "MatMul",
                    "BatchMatMul",
                    "BatchMatMulV2",
                    "Conv2DBackpropInput",
                    "Conv3DBackpropInputV2",
                ]
            ]
        )
        for i in target_nodes:
            if i[0] not in quantizable_op_names:
                continue
            computational_node_name = i[0]
            if self._ignore_insert_qdq_pattern(computational_node_name):
                continue

            computational_node = self.graph_info[computational_node_name].node
            weight_name = computational_node.input[1]
            if re.search(r"\w+:\d+", weight_name):
                weight_node = self.graph_info[weight_name.rsplit(":", 1)[0]].node
            else:
                weight_node = self.graph_info[weight_name].node
            if weight_node.op == "Enter":  # pragma: no cover
                if self.itex_mode:
                    parent_node = self.graph_info[Helper.node_name_from_input(weight_node.input[0])].node
                    if not parent_node.op == "Const":
                        continue
                    weight_node = parent_node
                else:
                    continue
            if computational_node_name in self.op_wise_config.keys():
                op_wise_cfg = self.op_wise_config[computational_node_name]
                per_channel = op_wise_cfg[0]
                weight_bit = op_wise_cfg[3]
            else:
                per_channel = False
                weight_bit = 7

            self._insert_qdq_pattern_for_weight_node(
                computational_node, weight_node, weight_name, min_max_values, per_channel, weight_bit, self.device
            )
        # Adaption for strip equivalent nodes feature
        # Replicate shared Dequantize for next step fusion
        self.g_qdq = GraphAnalyzer()
        self.g_qdq.graph = self.g_weight.dump_graph()
        self.graph_info = self.g_qdq.parse_graph()
        patterns = [["QuantizeV2"], ["Dequantize"]]
        matched_nodes = self.g_qdq.query_fusion_pattern_nodes(patterns)
        for i in matched_nodes:
            quantize_node_name = self.graph_info[i[0]].node.name
            deq_node_name = self.graph_info[i[1]].node.name
            deq_node = self.graph_info[i[1]].node
            len_deq_outputs = len(self.g_qdq.node_name_details[deq_node_name].outputs)
            if len_deq_outputs == 1:
                continue

            for index in range(len_deq_outputs - 1):
                rep_dequantize_node = Helper.create_node(
                    "Dequantize",
                    deq_node_name + "_" + str(index + 1),
                    [quantize_node_name, quantize_node_name + ":1", quantize_node_name + ":2"],
                )
                rep_dequantize_node.attr["T"].CopyFrom(deq_node.attr["T"])
                rep_dequantize_node.attr["mode"].CopyFrom(deq_node.attr["mode"])
                if "axis" in deq_node.attr:
                    rep_dequantize_node.attr["axis"].CopyFrom(deq_node.attr["axis"])
                next_node_name = self.g_qdq.node_name_details[deq_node_name].outputs[index + 1]
                self.g_qdq.add_node(rep_dequantize_node, quantize_node_name, [next_node_name])
                for input_index, each_input in enumerate(self.g_qdq.node_name_details[next_node_name].node.input):
                    if each_input == deq_node_name:
                        self.g_qdq.node_name_details[next_node_name].node.input[input_index] = rep_dequantize_node.name

        return self.g_qdq.dump_graph()

    def _check_op_list(self, node_type):
        """Check if the node_type in the allowed op list."""
        return any([node_type.find(i) != -1 for i in self.check_op_list])

    def _find_relu_node(self, node):
        """Find Relu node algorithm to identify the positive input."""
        if node.op == "MaxPool":
            self.check_op_list.add("BiasAdd")

        if (
            node.op in ("Relu", "Relu6", "Elu")
            or (
                node.op.find("AndRelu") != -1
                and ("alpha" not in node.attr or ("alpha" in node.attr and node.attr["alpha"].f == 0))
            )
        ) and (
            node.op != "Relu"
            or not self.performance_only
            or self.node_name_mapping[Helper.node_name_from_input(node.input[0])].node.op.find("FusedBatchNorm") == -1
            or self.node_name_mapping[Helper.node_name_from_input(node.input[0])].node.attr["is_training"].b
            or len(self.node_name_mapping[Helper.node_name_from_input(node.input[0])].output) > 1
        ):
            return True
        elif "T" in node.attr and dtypes.DType(node.attr["T"].type) in (dtypes.quint8, dtypes.uint8):
            return True
        elif (
            node.op.find("QuantizedConv") != -1
            or node.op.find("QuantizedDepthwiseConv") != -1
            or node.op.find("QuantizedMatMul") != -1
        ) and (
            (node.op.find("Relu") == -1 and node.op.find("Elu") == -1)
            or ("alpha" in node.attr and node.attr["alpha"].f > 0)
        ):
            return False
        elif self.itex_mode and node.op in ("Add", "AddV2", "AddN"):
            if re.search(r"\w+:\d+", node.input[0]):
                input0_node = self.node_name_mapping[node.input[0].rsplit(":", 1)[0]].node
            else:
                input0_node = self.node_name_mapping[node.input[0]].node
            if re.search(r"\w+:\d+", node.input[1]):
                input1_node = self.node_name_mapping[node.input[1].rsplit(":", 1)[0]].node
            else:
                input1_node = self.node_name_mapping[node.input[1]].node
            if input0_node.op in ("BiasAdd", "Add", "AddV2", "AddN") or input1_node.op in (
                "BiasAdd",
                "Add",
                "AddV2",
                "AddN",
            ):
                return False
            return self._find_relu_node(input0_node) and self._find_relu_node(input1_node)
        elif self._check_op_list(node.op) or (self.itex_mode and node.op in ("Add", "AddV2")):
            if node.op == "ConcatV2":
                find_relu = False
                for i in range(0, node.attr["N"].i):
                    if re.search(r"\w+:\d+", node.input[i]):
                        input_node = self.node_name_mapping[node.input[i].rsplit(":", 1)[0]].node
                    else:
                        input_node = self.node_name_mapping[node.input[i]].node
                    find_relu |= self._find_relu_node(input_node)
                return find_relu
            if re.search(r"\w+:\d+", node.input[0]):
                input_node = self.node_name_mapping[node.input[0].rsplit(":", 1)[0]].node
            else:
                input_node = self.node_name_mapping[node.input[0]].node
            return self._find_relu_node(input_node)
        else:
            return False

    def _insert_qdq_pattern_for_common_ops(self, original_node, is_asymmetric):
        """Insert QDQ patterns for common OPs."""
        namespace_prefix = original_node.name + "_eightbit"
        if original_node.op in ("Conv2DBackpropInput", "Conv3DBackpropInputV2"):
            all_inputs = self.node_name_mapping[original_node.name].node.input[-1:]
        else:
            all_inputs = self.node_name_mapping[original_node.name].node.input[:1]
        for each_input_name in all_inputs:
            if each_input_name[0] == "^":
                continue

            # if dq+maxpool is detected as input of this node
            # the qdq in pattern dq+maxpool+q should be with the same dtype in the itex mode
            if (
                self.itex_mode
                and each_input_name in self.node_name_mapping
                and self.node_name_mapping[each_input_name].node.op == "MaxPool"
                and self.graph_info[self.graph_info[each_input_name].node.input[0]].node.op == "Dequantize"
            ):
                maxpool_node = self.graph_info[each_input_name].node
                dtype = dtypes.DType(self.graph_info[maxpool_node.input[0]].node.attr["T"].type)
            elif self.node_name_mapping[original_node.name].node.op == "MatMul":
                dtype = dtypes.quint8
            elif (
                self.node_name_mapping[original_node.name].node.op == "BatchMatMulV2"
                or self.node_name_mapping[original_node.name].node.op == "BatchMatMul"
            ):
                dtype = dtypes.qint8
            # the qdq in pattern dq+bn+relu+q and dq+bn+q should be s8 in itex mode
            elif self.node_name_mapping[original_node.name].node.op == "FusedBatchNormV3":
                dtype = dtypes.qint8
            else:
                input_node_name = Helper.node_name_from_input(each_input_name)
                if input_node_name in self.graph_info:
                    if self.graph_info[input_node_name].node.op == "Dequantize":
                        dtype = dtypes.DType(self.graph_info[input_node_name].node.attr["T"].type)
                    elif self.graph_info[input_node_name].node.op == "FusedBatchNormV3":
                        dtype = dtypes.qint8
                    elif self._find_relu_node(self.node_name_mapping[original_node.name].node):
                        dtype = dtypes.quint8
                    else:
                        dtype = dtypes.qint8
                else:
                    dtype = (
                        dtypes.quint8
                        if self._find_relu_node(self.node_name_mapping[original_node.name].node)
                        else dtypes.qint8
                    )
            self._insert_qdq_pattern_for_each_input(
                original_node.name, namespace_prefix, each_input_name, is_asymmetric, dtype, device=self.device
            )

    def _insert_qdq_pattern_for_concatv2(self, original_node, is_asymmetric):
        """Insert QDQ patterns for each input of ConcatV2."""
        namespace_prefix = original_node.name + "_eightbit"
        normal_inputs = [i for i in original_node.input if i[0] != "^"]
        num_input = len(normal_inputs)
        original_inputs = normal_inputs[0 : num_input - 1]
        input_idx = 0
        for original_input_name in original_inputs:
            self._insert_qdq_pattern_for_each_input(
                original_node.name,
                namespace_prefix,
                original_input_name,
                is_asymmetric,
                dtypes.quint8,
                input_idx,
                device=self.device,
            )
            input_idx += 1

    def _insert_qdq_pattern_for_each_input(
        self, op_name, namespace_prefix, input_name, is_asymmetric, dtype=dtypes.quint8, input_index=0, device="cpu"
    ):
        """Takes one float input to an op, and converts it to quantized form."""
        unique_input_name = input_name.replace(":", "__port__").replace("^", "__hat__")
        min_input_name = namespace_prefix + "_min_" + unique_input_name
        max_input_name = namespace_prefix + "_max_" + unique_input_name
        quantize_input_name = namespace_prefix + "_quantize_" + unique_input_name

        reshape_dims_name = namespace_prefix + "_reshape_dims" + unique_input_name
        reduction_dims_name = namespace_prefix + "_reduction_dims" + unique_input_name

        if self.fake_quant:  # pragma: no cover
            min_node = Helper.create_constant_node(min_input_name, -1.0, dtypes.float32, device="cpu")
            max_node = Helper.create_constant_node(max_input_name, 1.0, dtypes.float32, device="cpu")
            quant_v2_node = Helper.create_node(
                "QuantizeV2", quantize_input_name, [input_name, min_input_name, max_input_name]
            )
            Helper.set_attr_dtype(quant_v2_node, "T", dtype)
            if not is_asymmetric:
                Helper.set_attr_string(quant_v2_node, "round_mode", b"HALF_TO_EVEN")
            # Helper.set_attr_bool(quant_v2_node, "narrow_range", False if is_asymmetric else True)
            if "BatchMatMul" in self.graph_info[op_name].node.op:
                Helper.set_attr_string(quant_v2_node, "mode", b"SCALED")
            else:
                Helper.set_attr_string(quant_v2_node, "mode", b"MIN_FIRST" if is_asymmetric else b"SCALED")

            if "Concat" in self.graph_info[op_name].node.op:
                dequantize_node = Helper.create_node(
                    "Dequantize",
                    op_name + "_dequantize_" + str(input_index),
                    [quant_v2_node.name, quant_v2_node.name + ":1", quant_v2_node.name + ":2"],
                )
            else:
                dequantize_node = Helper.create_node(
                    "Dequantize",
                    op_name + "_dequantize",
                    [quant_v2_node.name, quant_v2_node.name + ":1", quant_v2_node.name + ":2"],
                )
            Helper.set_attr_dtype(dequantize_node, "T", dtype)
            if "BatchMatMul" in self.graph_info[op_name].node.op:
                Helper.set_attr_string(dequantize_node, "mode", b"SCALED")
            else:
                Helper.set_attr_string(dequantize_node, "mode", b"MIN_FIRST" if is_asymmetric else b"SCALED")

            self.g.add_node(quant_v2_node, self.graph_info[op_name].node.input[0], [dequantize_node.name])
            self.g.add_node(dequantize_node, quant_v2_node.name, [op_name])
            self.g.add_node(min_node, None, [quant_v2_node.name])
            self.g.add_node(max_node, None, [quant_v2_node.name])
            self.graph_info[op_name].node.input[input_index] = dequantize_node.name
        else:
            reshape_dims_node = Helper.create_constant_node(reshape_dims_name, -1, dtypes.int32, [1])
            reduction_dims_node = Helper.create_constant_node(reduction_dims_name, 0, dtypes.int32, [1])
            reshape_input_name = namespace_prefix + "_reshape_" + unique_input_name
            if self.itex_mode and self.graph_info[op_name].node.op == "FusedBatchNormV3":
                min_input_name = namespace_prefix + "_input7_output_min"
                max_input_name = namespace_prefix + "_input8_output_max"
                quantize_input_name = namespace_prefix + "_quantize_bn"
            else:
                min_input_name = namespace_prefix + "_min_" + unique_input_name
                max_input_name = namespace_prefix + "_max_" + unique_input_name
                quantize_input_name = namespace_prefix + "_quantize_" + unique_input_name

            reshape_input_node = Helper.create_node("Reshape", reshape_input_name, [input_name, reshape_dims_name])
            Helper.set_attr_dtype(reshape_input_node, "T", dtypes.float32)

            min_input_node = Helper.create_node("Min", min_input_name, [reshape_input_name, reduction_dims_name])
            Helper.set_attr_dtype(min_input_node, "T", dtypes.float32)
            Helper.set_attr_dtype(min_input_node, "Tidx", dtypes.int32)
            Helper.set_attr_bool(min_input_node, "keep_dims", False)

            max_input_node = Helper.create_node("Max", max_input_name, [reshape_input_name, reduction_dims_name])
            Helper.set_attr_dtype(max_input_node, "T", dtypes.float32)
            Helper.set_attr_dtype(max_input_node, "Tidx", dtypes.int32)
            Helper.set_attr_bool(max_input_node, "keep_dims", False)

            if "BatchMatMul" in self.graph_info[op_name].node.op:
                min_input_node.input.append("^" + input_name)
                max_input_node.input.append("^" + input_name)

            if self.itex_mode:
                min_input_node.input.append("^" + input_name)
                max_input_node.input.append("^" + input_name)
            quant_v2_node = Helper.create_node(
                "QuantizeV2", quantize_input_name, [input_name, min_input_name, max_input_name]
            )
            Helper.set_attr_dtype(quant_v2_node, "T", dtype)
            if not is_asymmetric:
                Helper.set_attr_string(quant_v2_node, "round_mode", b"HALF_TO_EVEN")
            # Helper.set_attr_bool(quant_v2_node, "narrow_range", False if is_asymmetric else True)
            if self.performance_only or "BatchMatMul" in self.graph_info[op_name].node.op:
                Helper.set_attr_string(quant_v2_node, "mode", b"SCALED")
            else:
                Helper.set_attr_string(quant_v2_node, "mode", b"MIN_FIRST" if is_asymmetric else b"SCALED")

            if "Concat" in self.graph_info[op_name].node.op:
                dequantize_node = Helper.create_node(
                    "Dequantize",
                    op_name + "_dequantize_" + str(input_index),
                    [quant_v2_node.name, quant_v2_node.name + ":1", quant_v2_node.name + ":2"],
                )
            else:
                dequantize_node = Helper.create_node(
                    "Dequantize",
                    op_name + "_dequantize",
                    [quant_v2_node.name, quant_v2_node.name + ":1", quant_v2_node.name + ":2"],
                )
            Helper.set_attr_dtype(dequantize_node, "T", dtype)
            if self.performance_only or "BatchMatMul" in self.graph_info[op_name].node.op:
                Helper.set_attr_string(dequantize_node, "mode", b"SCALED")
            else:
                Helper.set_attr_string(dequantize_node, "mode", b"MIN_FIRST" if is_asymmetric else b"SCALED")
            if self.graph_info[op_name].node.op in ("Conv2DBackpropInput", "Conv3DBackpropInputV2"):
                input_index = 2

            self.g.add_node(quant_v2_node, self.graph_info[op_name].node.input[input_index], [dequantize_node.name])
            self.g.add_node(dequantize_node, quant_v2_node.name, [op_name])
            self.g.add_node(reshape_dims_node, None, [reshape_input_name])
            self.g.add_node(reduction_dims_node, None, [min_input_name, max_input_name])
            self.g.add_node(reshape_input_node, reshape_dims_name, [min_input_name, max_input_name])
            self.g.add_node(min_input_node, reshape_input_name, [quant_v2_node.name])
            self.g.add_node(max_input_node, reshape_input_name, [quant_v2_node.name])
            self.graph_info[op_name].node.input[input_index] = dequantize_node.name

    def _insert_qdq_pattern_for_weight_node(
        self, computational_node, weight_node, weight_name, min_max_values, per_channel, weight_bit=7.0, device="cpu"
    ):
        """Insert QDQ pattern for weight node."""
        host_op_type = computational_node.op
        base_name = weight_node.name + "_"
        qint8_const_name = base_name + "qint8_const"
        min_name = base_name + "min"
        max_name = base_name + "max"
        epsilon = 1e-4  # Needs to be set empirically if accuracy is not satisfactory
        range_coefficent = 127 / (2**weight_bit - 1)
        min_value = 0
        max_value = 0
        insert_reshape = False
        shape_convert = None
        shape_revert = None
        # The weight node of BatchMatMul may have no value
        if "value" in weight_node.attr and host_op_type in (
            "Conv2D",
            "MatMul",
            "BatchMatMul",
            "BatchMatMulV2",
            "Conv3D",
            "Conv2DBackpropInput",
            "Conv3DBackpropInputV2",
        ):
            float_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
            if per_channel:
                if host_op_type in ("Conv3D", "Conv3DBackpropInputV2"):
                    ranges = np.abs(float_tensor).max(axis=(0, 1, 2, 3))
                elif host_op_type in ("Conv2D", "Conv2DBackpropInput"):
                    ranges = np.abs(float_tensor).max(axis=(0, 1, 2))
                elif host_op_type in ("MatMul"):  # pragma: no cover
                    if "transpose_b" in weight_node.attr and weight_node.attr["transpose_b"].b:  # pragma: no cover
                        ranges = np.abs(float_tensor).max(axis=(1))
                    else:
                        # itex qdq needs to transpose this range
                        ranges = np.abs(float_tensor).max(axis=(0))
                else:
                    ranges = np.abs(float_tensor).max(axis=(0, 1))

                ranges *= range_coefficent
                min_value = -ranges
                max_value = ranges
                # nudging min-max values outside epsilon radius around zero
                ranges[ranges < epsilon] = epsilon
                min_value[np.abs(min_value) < epsilon] = -epsilon
                max_value[np.abs(max_value) < epsilon] = epsilon
                # qint8_tensor = (np.around(float_tensor *127.0/ranges)).astype(np.int8)
            else:
                min_value = np.min(float_tensor)
                max_value = np.max(float_tensor)
                min_value *= range_coefficent
                max_value *= range_coefficent
                min_value = min(min_value, 0.0)
                if min_value == max_value:
                    if abs(min_value) < 0.000001:
                        max_value = min_value + 1.0
                    elif min_value > 0:
                        max_value = 2 * min_value
                    else:
                        max_value = min_value / 2.0
                range_value = np.max(np.abs([min_value, max_value]))
                # qint8_tensor = (np.around(float_tensor * 127.0 / range_value)).astype(np.int8)
                # qint8_tensor = np.clip(qint8_tensor, -127, 127).astype(np.int8)
                min_value = -range_value
                max_value = range_value
        elif weight_node.op == "ReadVariableOp":
            min_value = self.llm_weight_minmax[weight_node.name][0]
            max_value = self.llm_weight_minmax[weight_node.name][1]
            min_value *= range_coefficent
            max_value *= range_coefficent
            min_value = min(min_value, 0.0)
            if min_value == max_value:
                if abs(min_value) < 0.000001:
                    max_value = min_value + 1.0
                elif min_value > 0:
                    max_value = 2 * min_value
                else:
                    max_value = min_value / 2.0
            range_value = np.max(np.abs([min_value, max_value]))
            # qint8_tensor = (np.around(float_tensor * 127.0 / range_value)).astype(np.int8)
            # qint8_tensor = np.clip(qint8_tensor, -127, 127).astype(np.int8)
            min_value = -range_value
            max_value = range_value
        elif host_op_type == "DepthwiseConv2dNative":
            float_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
            # get the max values based on dim 0 and 1 for depthwise conv
            # since, the output channel will be dim 2 * dim 3
            ranges = np.abs(float_tensor).max(axis=(0, 1))
            ranges = ranges.flatten()
            min_value = -ranges
            max_value = ranges
            # nudging min-max values outside epsilon radius around zero
            ranges[ranges < epsilon] = epsilon
            min_value[np.abs(min_value) < epsilon] = -epsilon
            max_value[np.abs(max_value) < epsilon] = epsilon
            # Since output channel will be 1 dim which is dim 2 * dim 3
            # When divide by range, qint8_tensor needs to be 3 dim
            # where, 3rd dim should be same dim of ranges
            a, b, c, d = float_tensor.shape
            # qint8_tensor = (np.around(float_tensor.reshape(a, b, c * d) * 127.0 /
            #                ranges)).astype(np.int8)
            # get the shape back to 4 dim
            # qint8_tensor = qint8_tensor.reshape(a, b, c, d)
            if self.itex_mode and d != 1:
                insert_reshape = True
                shape_convert = [a, b, c * d]
                shape_revert = [a, b, c, d]
        else:
            min_value = np.min(min_max_values[computational_node.name + "__min"])
            max_value = np.max(min_max_values[computational_node.name + "__max"])

        min_node = Helper.create_constant_node(min_name, min_value, dtypes.float32, device="cpu")
        max_node = Helper.create_constant_node(max_name, max_value, dtypes.float32, device="cpu")
        if "BatchMatMul" in host_op_type and "BatchMatMul" not in weight_node.op:
            min_node.input.append("^" + weight_name)
            max_node.input.append("^" + weight_name)

        min_enter_node = None
        max_enter_node = None
        if insert_reshape:
            reshape_dims_4to3_name = qint8_const_name + "_reshape_dims_4to3_"
            reshape_dims_4to3_node = Helper.create_constant_node(reshape_dims_4to3_name, shape_convert, dtypes.int32)
            reshape_4to3_name = qint8_const_name + "_reshape_4to3_"
            reshape_4to3_node = Helper.create_node(
                "Reshape", reshape_4to3_name, [weight_node.name, reshape_dims_4to3_name]
            )
            reshape_4to3_node.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
            quant_node = Helper.create_node(
                "QuantizeV2", qint8_const_name + "_quant", [reshape_4to3_name, min_name, max_name]
            )
        else:
            quant_node = Helper.create_node(
                "QuantizeV2", qint8_const_name + "_quant", [weight_node.name, min_name, max_name]
            )

        dequant_node = Helper.create_node(
            "Dequantize", base_name + "_dequant", [quant_node.name, quant_node.name + ":1", quant_node.name + ":2"]
        )
        Helper.set_attr_dtype(quant_node, "T", dtypes.qint8)
        Helper.set_attr_string(quant_node, "mode", b"SCALED")
        Helper.set_attr_string(quant_node, "round_mode", b"HALF_TO_EVEN")
        Helper.set_attr_dtype(dequant_node, "T", dtypes.qint8)
        Helper.set_attr_string(dequant_node, "mode", b"SCALED")
        if per_channel:
            if host_op_type in ("Conv2D", "Conv2DBackpropInput"):
                Helper.set_attr_int(quant_node, "axis", 3)
                Helper.set_attr_int(dequant_node, "axis", 3)
            elif host_op_type in ("Conv3D", "Conv3DBackpropInputV2"):
                Helper.set_attr_int(quant_node, "axis", 4)
                Helper.set_attr_int(dequant_node, "axis", 4)
            elif host_op_type == "MatMul":
                Helper.set_attr_int(quant_node, "axis", 1)
                Helper.set_attr_int(dequant_node, "axis", 1)
            else:
                Helper.set_attr_int(quant_node, "axis", -1)
                Helper.set_attr_int(dequant_node, "axis", -1)
        if host_op_type == "DepthwiseConv2dNative":
            Helper.set_attr_int(quant_node, "axis", 2)
            Helper.set_attr_int(dequant_node, "axis", 2)

        if insert_reshape:
            reshape_dims_3to4_name = qint8_const_name + "_reshape_dims_3to4_"
            reshape_dims_3to4_node = Helper.create_constant_node(reshape_dims_3to4_name, shape_revert, dtypes.int32)
            reshape_3to4_name = qint8_const_name + "_reshape_3to4_"
            reshape_3to4_node = Helper.create_node(
                "Reshape", reshape_3to4_name, [dequant_node.name, reshape_dims_3to4_name]
            )
            reshape_3to4_node.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
            self.g_weight.add_node(reshape_dims_4to3_node, None, [reshape_4to3_name])
            self.g_weight.add_node(reshape_dims_3to4_node, None, [reshape_3to4_name])
            self.g_weight.add_node(reshape_4to3_node, weight_node.name, [quant_node.name])
            self.g_weight.add_node(quant_node, reshape_4to3_name, [])
            self.g_weight.add_node(min_node, None, [quant_node.name])
            self.g_weight.add_node(max_node, None, [quant_node.name])
            self.g_weight.add_node(dequant_node, quant_node.name, [reshape_3to4_name])
            self.g_weight.add_node(reshape_3to4_node, dequant_node.name, [computational_node.name])
            computational_node.input[1] = reshape_3to4_node.name
        else:
            if (
                computational_node.name in self.g.parent_frame_details
                and self.g.parent_frame_details[computational_node.name]
            ):  # pragma: no cover
                weight_enter_node = Helper.create_node("Enter", weight_node.name + "_enter", [weight_node.name])
                Helper.set_attr_string(
                    weight_enter_node,
                    "frame_name",
                    self.g.parent_frame_details[computational_node.name].attr["frame_name"].s,
                )
                Helper.set_attr_dtype(weight_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(weight_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    weight_enter_node,
                    "parallel_iterations",
                    self.g.parent_frame_details[computational_node.name].attr["parallel_iterations"].i,
                )

                min_enter_node = Helper.create_node("Enter", min_name + "_enter", [min_name])
                Helper.set_attr_string(
                    min_enter_node,
                    "frame_name",
                    self.g.parent_frame_details[computational_node.name].attr["frame_name"].s,
                )
                Helper.set_attr_dtype(min_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(min_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    min_enter_node,
                    "parallel_iterations",
                    self.g.parent_frame_details[computational_node.name].attr["parallel_iterations"].i,
                )

                max_enter_node = Helper.create_node("Enter", max_name + "_enter", [max_name])
                Helper.set_attr_string(
                    max_enter_node,
                    "frame_name",
                    self.g.parent_frame_details[computational_node.name].attr["frame_name"].s,
                )
                Helper.set_attr_dtype(max_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(max_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    max_enter_node,
                    "parallel_iterations",
                    self.g.parent_frame_details[computational_node.name].attr["parallel_iterations"].i,
                )

                self.g_weight.add_node(quant_node, weight_name, [])
                self.g_weight.add_node(min_node, None, [min_enter_node.name])
                self.g_weight.add_node(max_node, None, [max_enter_node.name])
                self.g_weight.add_node(min_enter_node, min_node.name, [quant_node.name])
                self.g_weight.add_node(max_enter_node, max_node.name, [quant_node.name])
                self.g_weight.add_node(weight_enter_node, weight_node.name, [quant_node.name])
                quant_node.input[0] = weight_enter_node.name
                quant_node.input[1] = min_enter_node.name
                quant_node.input[2] = max_enter_node.name
                self.g_weight.add_node(quant_node, weight_enter_node.name, [])
                self.g_weight.add_node(dequant_node, quant_node.name, [computational_node.name])
                computational_node.input[1] = dequant_node.name
            else:
                self.g_weight.add_node(quant_node, weight_name, [])
                self.g_weight.add_node(min_node, None, [quant_node.name])
                self.g_weight.add_node(max_node, None, [quant_node.name])
                self.g_weight.add_node(dequant_node, quant_node.name, [computational_node.name])
                computational_node.input[1] = dequant_node.name

    def _ignore_insert_qdq_pattern(self, matched_node_name):
        """For some cases we don't need to insert QDQ patterns."""
        if (matched_node_name in self.fp32_ops or matched_node_name in self.bf16_ops) and (
            (matched_node_name,) not in self.quantized_nodes
        ):
            return True

        if matched_node_name not in self.op_wise_config and (matched_node_name,) not in self.quantized_nodes:
            return True

        # TODO Remove below two lines once the TF enabled the QuantizedMatMul while
        # transpose_a could be set to True.
        if not self.itex_mode and self.graph_info[matched_node_name].node.op == "MatMul":
            if self.graph_info[matched_node_name].node.attr["transpose_a"].b is True:
                return True
        if "FusedBatchNorm" in self.graph_info[matched_node_name].node.op and not self.itex_mode:
            return True
        if "_MklFusedInstanceNorm" == self.graph_info[matched_node_name].node.op and not self.itex_mode:
            return True
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""All frameworks-agnostic constants."""

# constants for configs
GLOBAL = "global"
LOCAL = "local"
DEFAULT_WHITE_LIST = "*"
EMPTY_WHITE_LIST = None

# config name
BASE_CONFIG = "base_config"
COMPOSABLE_CONFIG = "composable_config"
RTN = "rtn"
STATIC_QUANT = "static_quant"
SMOOTH_QUANT = "smooth_quant"
GPTQ = "gptq"
AWQ = "awq"  # pragma: no cover
HQQ = "hqq"  # pragma: no cover
TEQ = "teq"  # pragma: no cover
AUTOROUND = "autoround"
FP8_QUANT = "fp8_quant"
MX_QUANT = "mx_quant"
MIXED_PRECISION = "mixed_precision"

# options
import datetime

DEFAULT_WORKSPACE = "./nc_workspace/{}/".format(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))

from typing import Callable, Union

OP_NAME_OR_MODULE_TYPE = Union[str, Callable]

# mode name
from enum import Enum


class Mode(Enum):
    """Enumeration class representing different modes of the quantizer execution."""

    PREPARE = "prepare"
    CONVERT = "convert"
    QUANTIZE = "quantize"


SERVER_PROCESSOR_BRAND_KEY_WORLD_LST = ["Xeon"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Meta OP Graph Rewriter."""


from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class MetaInfoChangingMemOpOptimizer(GraphRewriterBase):
    """Fuse the pattern like Dequantize + MetaOp + Quantize into MetaOp(set its type to int8).

    With such changes, the Quantize and Dequantize OP will removed for better performance.
    """

    def __init__(self, model):
        """Initialization."""
        super().__init__(model)

        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

    @dump_elapsed_time("Pass MetaOpOptimizer")
    def do_transformation(self):
        """Apply the fusion of Dequantize + MetaOp + QuantizeV2."""
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(
            [["Dequantize"], ("Squeeze", "Reshape"), ("Squeeze", "Reshape"), ["QuantizeV2"]]
        )
        for i in target_nodes:
            if len(i[-1]) == 2:
                continue

            dequantize_node_name = i[0]
            if len(i[-1]) == 3:
                quantize_node_name = i[2]
            else:
                quantize_node_name = i[3]
                dequantize_node_name = i[0]
                if self.graph_info[i[1]].node.op == self.graph_info[i[2]].node.op:
                    continue
            deq_node = self.graph_info[dequantize_node_name].node
            quant_node = self.graph_info[quantize_node_name].node
            if len(self.graph_info[dequantize_node_name].outputs) != 1:
                continue

            if quant_node.attr["mode"].s.decode() == deq_node.attr["mode"].s.decode():
                deq_min_range = self.graph_info[dequantize_node_name].node.input[1]
                deq_max_range = self.graph_info[dequantize_node_name].node.input[2]
                quant_output_min = quantize_node_name + ":1"
                quant_output_max = quantize_node_name + ":2"
                if len(i[-1]) == 3:
                    quantize_input_name = i[1]
                else:
                    quantize_input_name = i[2]

                quantized_node_name = self.graph_info[quantize_node_name].outputs[0]
                # _QuantizedBatchMatMul requires T1 and T2 with qint8 type
                # _QuantizedFusedBatchNorm requires T with qint8 type
                if (
                    self.graph_info[quantized_node_name].node.op == "_QuantizedBatchMatMul"
                    or self.graph_info[quantized_node_name].node.op == "_QuantizedFusedBatchNorm"
                ) and self.graph_info[dequantize_node_name].node.attr["T"].type != dtypes.qint8.as_datatype_enum:
                    continue

                for index, value in enumerate(self.graph_info[quantized_node_name].node.input):
                    if value == quant_output_min:
                        self.graph_info[quantized_node_name].node.input[index] = deq_min_range

                    if value == quant_output_max:
                        self.graph_info[quantized_node_name].node.input[index] = deq_max_range

                    if index == 0:
                        self.graph_info[quantized_node_name].node.input[index] = quantize_input_name

                new_dtype = self.graph_info[dequantize_node_name].node.attr["T"].type
                for node_name in i[1:-1]:
                    self.graph_info[node_name].node.attr["T"].type = new_dtype

                if "T1" in self.graph_info[quantized_node_name].node.attr:
                    self.graph_info[quantized_node_name].node.attr["T1"].type = new_dtype

                if "Tinput" in self.graph_info[quantized_node_name].node.attr:
                    self.graph_info[quantized_node_name].node.attr["Tinput"].type = new_dtype

                if "Thost_inputs" in self.graph_info[quantized_node_name].node.attr:
                    self.graph_info[quantized_node_name].node.attr["Thost_inputs"].list.type[0] = new_dtype

                if "T" in self.graph_info[quantized_node_name].node.attr:
                    self.graph_info[quantized_node_name].node.attr["T"].type = new_dtype

                self.graph_info[i[1]].node.input[0] = self.graph_info[dequantize_node_name].node.input[0]
                self.graph_analyzer.remove_node(dequantize_node_name)
                self.graph_analyzer.remove_node(self.graph_info[quantize_node_name].node.input[1])
                self.graph_analyzer.remove_node(self.graph_info[quantize_node_name].node.input[2])
                self.graph_analyzer.remove_node(quantize_node_name)
        return GraphAnalyzer().dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dilated Contraction Graph Rewriter."""

from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class DilatedContraction(GraphRewriterBase):
    """Fuse the SpaceToBatchND + Conv + BatchToSpaceND pattern."""

    @dump_elapsed_time("Pass DilatedContraction")
    def do_transformation(self):
        """Dilated Contraction fusion."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        target_nodes = cur_graph.query_fusion_pattern_nodes(
            ["SpaceToBatchND", ["Conv2D", "DepthwiseConv2dNative"], "BatchToSpaceND"]
        )

        for node_combination in target_nodes:
            stob_node = graph_info[node_combination[0]].node
            contraction_node = graph_info[node_combination[1]].node
            btos_node = graph_info[node_combination[2]].node
            stob_padding_node = graph_info[stob_node.input[2]].node

            block_shape_node = graph_info[btos_node.input[1]].node
            crops_node = graph_info[btos_node.input[2]].node

            block_value = [i for i in tensor_util.MakeNdarray(block_shape_node.attr["value"].tensor).flat]
            new_dilation = [1, block_value[0], block_value[1], 1]
            # if padding input of SpaceToBatchND can't be directly fetched, we continue
            if stob_padding_node.op != "Const":
                continue
            padding_value = [i for i in tensor_util.MakeNdarray(stob_padding_node.attr["value"].tensor).flat]
            crops_value = [i for i in tensor_util.MakeNdarray(crops_node.attr["value"].tensor).flat]

            contraction_node.input[0] = stob_node.input[0]
            Helper.set_attr_int_list(contraction_node, "dilations", new_dilation)

            real_padding = [padding_value[i] - crops_value[i] for i in range(4)]
            explict_padding = [0, 0, 0, 0, 0, 0, 0, 0]
            data_format = contraction_node.attr["data_format"].s.decode()
            if any(real_padding):
                contraction_node.attr["padding"].s = "EXPLICIT".encode()
                assert data_format in ("NHWC", "NCHW")
                if data_format == "NHWC":
                    explict_padding[2] = real_padding[0]
                    explict_padding[3] = real_padding[1]
                    explict_padding[4] = real_padding[2]
                    explict_padding[5] = real_padding[3]
                else:
                    explict_padding[4] = real_padding[0]
                    explict_padding[5] = real_padding[1]
                    explict_padding[6] = real_padding[2]
                    explict_padding[7] = real_padding[3]
                Helper.set_attr_int_list(contraction_node, "explicit_paddings", explict_padding)

            contraction_node.attr.pop("_output_shapes")
            cur_graph.remove_node(stob_node.name)
            following_node_name = graph_info[node_combination[2]].outputs[0]
            following_node = graph_info[following_node_name].node

            following_node.input[0] = btos_node.input[0]
            cur_graph.remove_node(btos_node.name)

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Move Squeeze after Relu Graph Rewriter."""

import copy

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class MoveSqueezeAfterReluOptimizer(GraphRewriterBase):
    """Move Squeeze op after Relu op for match fusion pattern."""

    def __init__(self, model):
        """Initialization."""
        super().__init__(model)
        self.op_list = ["Relu", "Sigmoid", "Relu6", "LeakyRelu", "Elu"]

    @dump_elapsed_time("Pass MoveSqueezeAfterReluOptimizer")
    def do_transformation(self):
        """Move Squeeze/Reshape after Relu."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()
        # For pattern Conv + Squeeze + BiasAdd + Relu(Sigmoid, Relu6, LeakyRelu, Elu)
        for node in self.model.node:
            if (
                node.op in self.op_list
                and node.input[0] in graph_info
                and graph_info[node.input[0]].node.op == "BiasAdd"
            ):
                biasadd_node = graph_info[node.input[0]].node
                biasadd_input = graph_info[biasadd_node.name].node.input[0]
                squeeze_node = graph_info[biasadd_input].node
                relu_output = graph_info[node.name].outputs
                if squeeze_node.op == "Squeeze":
                    # biasadd
                    for i, input in enumerate(biasadd_node.input):
                        if input == biasadd_input:
                            new_input = biasadd_node.input[:i] + [squeeze_node.input[0]] + biasadd_node.input[i + 1 :]
                            graph_info[biasadd_node.name].node.ClearField("input")
                            graph_info[biasadd_node.name].node.input.extend(new_input)
                            graph_info[squeeze_node.name].outputs.remove(biasadd_node.name)
                    # conv output
                    conv = squeeze_node.input[0]
                    conv_outputs = graph_info[conv].outputs
                    for i, output in enumerate(conv_outputs):
                        if output == squeeze_node.name:
                            graph_info[conv].outputs.remove(squeeze_node.name)
                            graph_info[conv].outputs.append(biasadd_node.name)
                    # squeeze input
                    squeeze_node.ClearField("input")
                    squeeze_node.input.extend([node.name])
                    # expand input,squeeze output
                    for output in relu_output:
                        for i, input in enumerate(graph_info[output].node.input):
                            if input == node.name:
                                new_input = (
                                    graph_info[output].node.input[:i]
                                    + [squeeze_node.name]
                                    + graph_info[output].node.input[i + 1 :]
                                )
                                graph_info[squeeze_node.name].outputs.append(output)
                                graph_info[output].node.ClearField("input")
                                graph_info[output].node.input.extend(new_input)

            # For pattern x + Reshape + Relu(Sigmoid, Relu6, LeakyRelu, Elu)
            if (
                node.op in self.op_list
                and node.input[0] in graph_info
                and graph_info[node.input[0]].node.op == "Reshape"
            ):
                reshape_node = graph_info[node.input[0]].node
                reshape_input = graph_info[reshape_node.name].node.input[0]
                x_node = graph_info[reshape_input].node
                relu_output = copy.deepcopy(graph_info[node.name].outputs)

                if len(graph_info[x_node.name].outputs) != 1:
                    continue
                if len(graph_info[reshape_node.name].outputs) > 1:
                    continue
                # relu---->reshape
                for i, input in enumerate(reshape_node.input):
                    if input == reshape_input:
                        new_input = reshape_node.input[:i] + [node.name] + reshape_node.input[i + 1 :]
                        graph_info[reshape_node.name].node.ClearField("input")
                        graph_info[reshape_node.name].node.input.extend(new_input)
                        graph_info[x_node.name].outputs.remove(reshape_node.name)
                        graph_info[x_node.name].outputs.append(node.name)
                # x----->relu
                node.ClearField("input")
                node.input.extend([reshape_input])
                # expand input,squeeze output
                for output in relu_output:
                    for i, input in enumerate(graph_info[output].node.input):
                        if input == node.name:
                            new_input = (
                                graph_info[output].node.input[:i]
                                + [reshape_node.name]
                                + graph_info[output].node.input[i + 1 :]
                            )
                            graph_info[reshape_node.name].outputs.append(output)
                            graph_info[output].node.ClearField("input")
                            graph_info[output].node.input.extend(new_input)
                            graph_info[node.name].outputs.remove(output)
                            graph_info[node.name].outputs.append(reshape_node.name)
        return g.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Post HostConst Graph Rewriter."""

import os

from tensorflow.core.framework import graph_pb2, node_def_pb2

from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class PostHostConstConverter(GraphRewriterBase):
    """Support HostConst as default for all devices, not just for GPU."""

    @dump_elapsed_time("Pass PostHostConstConverter")
    def do_transformation(self):
        """Convert Const to HostConst as default."""
        if os.environ.get("DISABLE_HOSTCONST") == "1":
            return self.model
        output_graph_def = graph_pb2.GraphDef()
        for node in self.model.node:
            new_node = node_def_pb2.NodeDef()
            new_node.CopyFrom(node)
            new_node.device = ""
            if (
                node.op == "Const"
                and node.attr["dtype"].type in [1, 3]
                and (
                    node.name.endswith("_min")
                    or node.name.endswith("_max")
                    or node.name.endswith("_max_only")
                    or node.name.endswith("_min_only")
                )
            ):
                new_node.op = "HostConst"
            output_graph_def.node.extend([new_node])
        return output_graph_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Scale propagation Graph Rewriter."""

from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from ..graph_base import GraphRewriterBase


class ScaleProPagationTransformer(GraphRewriterBase):
    """Scale propagation converter."""

    cac_pattern = [
        ["QuantizeV2", "Requantize", "RequantizePerChannel"],
        ["QuantizedAvgPool"],
        [
            "QuantizedConv2DWithBias",
            "QuantizedConv2DWithBiasAndRelu",
            "QuantizedConv2DPerChannel",
            "QuantizedConv2D",
            "QuantizedConv2DWithBiasSumAndRelu",
        ],
        ["Requantize", "RequantizePerChannel"],
    ]

    def __init__(self, model, direction="Up"):
        """Initialization."""
        super().__init__(model)
        self.direction = direction if direction not in ("Up", "Down") else "Up"
        self.cur_graph = GraphAnalyzer()
        self.cur_graph.graph = self.model

        self.graph_info = self.cur_graph.parse_graph()

    def _create_new_const_node(self, new_const_node_name, value, old_const_node_name):
        """Create new Const node."""
        new_node = node_def_pb2.NodeDef()
        new_node.op = "Const"
        new_node.name = new_const_node_name
        new_node.attr["dtype"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
        new_node.attr["value"].CopyFrom(
            attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(float(value), dtypes.float32, []))
        )
        output_node_name = self.graph_info[old_const_node_name].outputs[0]
        self.cur_graph.replace_const_node(
            new_node, [Helper.node_name_from_input(output_node_name)], old_const_node_name
        )
        self.cur_graph.remove_node(old_const_node_name)

    def _cac_transformation(self):
        """Scale propagation pattern match and create new const nodes."""
        target_nodes = self.cur_graph.query_fusion_pattern_nodes(self.cac_pattern)
        quantize_v2_min_index = 1
        requntize_min_index = 3
        quantize_v2_max_index = 2
        requntize_max_index = 4

        for match in target_nodes:
            pre_node_name = match[0]

            pre_node = self.graph_info[pre_node_name].node

            output_nodes_count = len(set(self.graph_info[pre_node_name].outputs))

            if output_nodes_count > 1:
                continue
            # Skip transformation if avgpool has multi output nodes.
            pooling_nodes_count = len(set(self.graph_info[match[1]].outputs))
            if pooling_nodes_count > 1:
                continue

            if pre_node.op == "QuantizeV2":
                pre_min_index, pre_max_index = quantize_v2_min_index, quantize_v2_max_index
            else:
                pre_min_index, pre_max_index = requntize_min_index, requntize_max_index

            requantize_node_name = match[3]
            requantize_node = self.graph_info[requantize_node_name].node

            requantize_min = self.graph_info[
                Helper.node_name_from_input(requantize_node.input[requntize_min_index])
            ].node
            requantize_max = self.graph_info[
                Helper.node_name_from_input(requantize_node.input[requntize_max_index])
            ].node

            requantize_min_value = (requantize_min.attr["value"].tensor.float_val)[0]
            requantize_max_value = (requantize_max.attr["value"].tensor.float_val)[0]
            self._create_new_const_node(
                pre_node_name + "_cac_requantize_min_value", requantize_min_value, pre_node.input[pre_min_index]
            )
            self._create_new_const_node(
                pre_node_name + "_cac_requantize_max_value", requantize_max_value, pre_node.input[pre_max_index]
            )

    def do_transformation(self):
        """Apply the scale propagation algrothim."""
        self._cac_transformation()

        return self.cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initialize custom layers for Keras quantization."""

from neural_compressor.tensorflow.keras.layers.conv2d import QConv2D
from neural_compressor.tensorflow.keras.layers.dense import QDense
from neural_compressor.tensorflow.keras.layers.depthwise_conv2d import QDepthwiseConv2D
from neural_compressor.tensorflow.keras.layers.pool2d import QAvgPool2D, QMaxPool2D
from neural_compressor.tensorflow.keras.layers.separable_conv2d import QSeparableConv2D
from neural_compressor.tensorflow.keras.layers.layer_initializer import layer_initializer_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize the AvgPool/MaxPool."""

import tensorflow as tf
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper
from neural_compressor.tensorflow.utils import version1_eq_version2, version1_gt_version2, version1_lt_version2

from ..quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithPooling(QuantizeNodeBase):
    """Quantize the AvgPool and MaxPool."""

    def __init__(self, **kwargs):
        """Initialization."""
        super().__init__(**kwargs)
        self.sorted_patterns = sorted(self.patterns, key=lambda i: len(i), reverse=True)

    def _add_pool_function(self, original_node, quantized_op_node):
        """Set quantized pooling node attributes."""
        pooling_type = (
            dtypes.quint8
            if version1_lt_version2(tf.version.VERSION, "2.6.0") or self._find_relu_node(original_node)
            else dtypes.qint8
        )
        helper.set_attr_dtype(quantized_op_node, "T", pooling_type)
        helper.copy_attr(quantized_op_node, "ksize", original_node.attr["ksize"])
        helper.copy_attr(quantized_op_node, "strides", original_node.attr["strides"])
        helper.copy_attr(quantized_op_node, "padding", original_node.attr["padding"])

    def _apply_pool_quantization(self, match_node_name):
        """Quantize AvgPool/MaxPool.

        Dequantize + MaxPool + QuantizeV2
        Dequantize + MaxPool3D + QuantizeV2
        Dequantize + AvgPool + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        _, q_inputs = self._get_node_input(normal_inputs[0])

        all_input_names = q_inputs
        skip_node_name.append(normal_inputs[0])

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))
                quantized_op_name = node.name + "_eightbit_quantized"
                quantized_op_type = "Quantized" + node.op
                if node.op == "MaxPool3D":
                    quantized_op_type = "_Quantized" + node.op

                quantized_pool_node = helper.create_node(quantized_op_type, quantized_op_name, all_input_names)

                self._add_pool_function(node, quantized_pool_node)
                self.add_output_graph_node(quantized_pool_node)
                deq_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                self._intel_cpu_add_dequantize_result_node(
                    quantized_op_name, node.name, dtype=deq_type, performance_only=self.performance_only
                )
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        """Get the longest fusion pattern."""
        self._get_op_list()
        matched_node_name = []

        for k, v in enumerate(self.op_list):
            if v in set(fusion[1] for fusion in self.sorted_patterns):
                cur_node = self.node_name_mapping[list(self.node_name_mapping.keys())[k]].node

                if cur_node.name != self.start_node_name:
                    continue

                for sub_rule in self.sorted_patterns:
                    if sub_rule[0] != "Dequantize" or sub_rule[-1] != "QuantizeV2":
                        continue
                    if v != sub_rule[1]:
                        continue
                    matched_node_name.clear()
                    matched_node_name.append(sub_rule[0])
                    matched_node_name.append(cur_node.name)
                    matched_node_name.append(sub_rule[-1])
                    return sub_rule, matched_node_name
        return None, None

    def apply_the_transform(self):
        """Quantize AvgPool/MaxPool."""
        self._get_op_list()
        matched_rule, matched_node_name = self.get_longest_fuse()
        if matched_node_name:
            fusion_name = "".join(matched_rule)
            if (
                fusion_name == "DequantizeMaxPoolQuantizeV2"
                or fusion_name == "DequantizeMaxPool3DQuantizeV2"
                or fusion_name == "DequantizeAvgPoolQuantizeV2"
            ):
                self._apply_pool_quantization(matched_node_name)
            else:  # pragma: no cover
                self.logger.info("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, []

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)
            return self.output_graph, []

        if self.remove_redundant_quant_flag:
            self.input_graph = self.remove_redundant_quantization(self.input_graph)
        return self.input_graph, []
